from typing import TYPE_CHECKING

from django.db import models
from django.db.transaction import atomic

from django_napse.core.models.histories.wallet import WalletHistory

if TYPE_CHECKING:
    from django_napse.core.models.transactions.credit import Credit
    from django_napse.core.models.wallets.wallet import Wallet


class CreditManager(models.Manager):
    """Manager for the Credit model."""

    @atomic()
    def create(self, wallet: "Wallet", amount: float, ticker: str) -> "Credit":
        """Create a Credit object.

        Args:
            wallet (Wallet): The wallet to credit.
            amount (float): The amount to credit.
            ticker (str): The ticker of the currency to credit.

        Returns:
            Credit: The created credit.
        """
        if amount == 0:
            return None
        credit = self.model(
            wallet=wallet,
            amount=amount,
            ticker=ticker,
        )

        credit.save()
        wallet.top_up(amount, ticker, force=True)
        WalletHistory.get_or_create(wallet).generate_data_point()
        return credit



from typing import Optional

from django.db import models


class ArchitectureManager(models.Manager):
    def create(self, constants: Optional[dict] = None, variables: Optional[dict] = None) -> models.Model:
        constants = constants or {}
        variables = variables or {}
        architecture = self.model(**constants, **{f"variable_{key}": value for key, value in variables.items()})
        architecture.save()
        return architecture



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Optional

from django.db import models

from django_napse.core.models.connections.connection import Connection
from django_napse.core.models.modifications import ArchitectureModification, ConnectionModification, StrategyModification
from django_napse.core.models.orders.order import Order, OrderBatch
from django_napse.utils.errors import BotError

if TYPE_CHECKING:
    from django_napse.core.models.wallets.space_simulation_wallet import SpaceSimulationWallet
    from django_napse.core.models.wallets.space_wallet import SpaceWallet


class Bot(models.Model):
    uuid = models.UUIDField(unique=True, editable=False, default=uuid.uuid4)

    name = models.CharField(max_length=100, default="Bot")

    created_at = models.DateTimeField(auto_now_add=True)
    active = models.BooleanField(default=True)

    strategy = models.OneToOneField("Strategy", on_delete=models.CASCADE, related_name="bot")

    def __str__(self) -> str:
        return f"BOT {self.pk=}"

    def save(self, *args: list, **kwargs: dict) -> None:  # noqa: D102
        if self.is_in_simulation and self.is_in_fleet:
            error_msg = "Bot is in simulation and fleet."
            raise BotError.InvalidSetting(error_msg)
        return super().save(*args, **kwargs)

    def info(self, beacon: str = "", *, verbose: bool = True) -> str:
        """Return a string with the model information.

        Args:
            beacon (str, optional): The prefix for each line. Defaults to "".
            verbose (bool, optional): Whether to print the string. Defaults to True.

        Returns:
            str: The string with the history information.
        """
        string = ""
        string += f"{beacon}Bot {self.pk}:\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.name=}\n"

        string += f"{beacon}Strategy:\n"
        new_beacon = beacon + "\t"
        string += f"{beacon}\t{self.strategy.info(verbose=False, beacon=new_beacon)}\n"

        if verbose:  # pragma: no cover
            print(string)
        return string

    @property
    def is_in_simulation(self):
        return hasattr(self, "simulation")

    @property
    def is_in_fleet(self):
        # return hasattr(self, "bot_in_cluster")
        return hasattr(self, "link")

    @property
    def is_templace(self):
        """Is self a template bot of a cluster?"""
        return hasattr(self, "cluster")

    @property
    def is_free(self):
        """Is self not in a simulation, not in a fleet and not a cluster's template bot?"""
        return not self.is_in_simulation and not self.is_in_fleet and not self.is_templace

    @property
    def testing(self):
        if self.is_in_simulation:
            return True
        if self.is_in_fleet:
            return self.bot_in_cluster.cluster.fleet.testing
        error_msg = "Bot is not in simulation or fleet."
        raise BotError.InvalidSetting(error_msg)

    @property
    def fleet(self):
        if self.is_in_fleet:
            return self.link.cluster.fleet
        return None

    @property
    def space(self):
        if self.is_in_simulation:
            return self.simulation.space
        if self.is_in_fleet:
            error_msg = "Bot is in a fleet and therefore doesn't have a space."
            raise BotError.NoSpace(error_msg)
        error_msg = "Bot is not in simulation or fleet."
        raise BotError.InvalidSetting(error_msg)

    @property
    def exchange_account(self):
        if self.is_in_simulation:
            return self.simulation.space.exchange_account
        if self.is_in_fleet:
            return self.link.cluster.fleet.exchange_account
        error_msg = "Bot is not in simulation or fleet."
        raise BotError.InvalidSetting(error_msg)

    @property
    def _strategy(self):
        return self.strategy.find()

    @property
    def architecture(self):
        return self._strategy.architecture.find()

    @property
    def controllers(self):
        return self.architecture.controllers_dict()

    @property
    def orders(self):
        connections = self.connections.select_related("orders").all()
        return [connection.orders for connection in connections]

    def hibernate(self):
        if not self.active:
            error_msg = "Bot is already hibernating."
            raise BotError.InvalidSetting(error_msg)
        self.active = False
        self.save()

    def get_connections(self):
        return list(self.connections.all())

    def get_connection_data(self):
        return {connection: connection.to_dict() for connection in self.get_connections()}

    def get_orders(self, data: Optional[dict] = None, no_db_data: Optional[dict] = None):
        if not self.active:
            error_msg = "Bot is hibernating."
            raise BotError.InvalidSetting(error_msg)

        orders = self._get_orders(data=data, no_db_data=no_db_data)
        batches = {}
        order_objects = []
        for order in orders:
            controller = order["controller"]
            batches[controller] = OrderBatch.objects.create(controller=controller)
        for order_dict in orders:
            controller = order_dict.pop("controller")
            strategy_modifications = order_dict.pop("StrategyModifications")
            connection_modifications = order_dict.pop("ConnectionModifications")
            architecture_modifications = order_dict.pop("ArchitectureModifications")
            order = Order.objects.create(batch=batches[controller], **order)
            order_objects.append(order)
            for modification in strategy_modifications:
                StrategyModification.objects.create(order=order, **modification)
            for modification in connection_modifications:
                ConnectionModification.objects.create(order=order, **modification)
            for modification in architecture_modifications:
                ArchitectureModification.objects.create(order=order, **modification)
        for batch in batches.values():
            batch.set_status_ready()

        return order_objects, batches

    def _get_orders(self, data: Optional[dict] = None, no_db_data: Optional[dict] = None):
        """Get orders of the bot."""
        return self.architecture._get_orders(data=data, no_db_data=no_db_data)  # noqa: SLF001

    def connect_to_wallet(self, wallet: SpaceSimulationWallet | SpaceWallet) -> Connection:
        """Connect the bot to a (sim)space's wallet."""
        connection = Connection.objects.create(owner=wallet, bot=self)
        for plugin in self._strategy.plugins.all():
            plugin.connect(connection)
        self._strategy.connect(connection)
        return connection

    def copy(self) -> Bot:
        """Copy the instance bot."""
        return self.__class__.objects.create(
            name=f"Copy of {self.name}",
            strategy=self.strategy.copy(),
        )

    def value(self, space=None) -> float:  # noqa: ANN001
        """Return value market of the bot, depending of space containerization."""
        if space is None:
            return sum([connection.wallet.value_market() for connection in self.connections.all()])
        connection = Connection.objects.get(owner=space.wallet, bot=self)
        return connection.wallet.value_market()

    def get_stats(self, space=None) -> dict[str, str | int | float]:  # noqa: ANN001
        """Some bot's statistics used for KPI dashboards."""
        return {
            "value": self.value(space),
            "profit": 0,
            "delta_30": 0,  # TODO: Need history
        }



from datetime import timedelta

from django.core.management.base import BaseCommand

from django_napse.core.models import Bot, Controller, DCABotConfig, DCAStrategy, ExchangeAccount, SinglePairArchitecture, Space


class Command(BaseCommand):  # noqa
    help = "Closes the specified poll for voting"

    def add_arguments(self, parser):  # noqa
        pass

    def handle(self, *args, **options):  # noqa
        exchange_account = ExchangeAccount.objects.first()
        space = Space.objects.first()
        config = DCABotConfig.objects.create(space=space, settings={"timeframe": timedelta(hours=1)})
        controller = Controller.get(
            exchange_account=exchange_account,
            base="BTC",
            quote="USDT",
            interval="1m",
        )
        architecture = SinglePairArchitecture.objects.create(constants={"controller": controller})
        strategy = DCAStrategy.objects.create(config=config, architecture=architecture)
        Bot.objects.create(name="Test Bot", strategy=strategy)

        self.stdout.write(self.style.SUCCESS("DCABot have been created successfully!"))



import shutil
import time
from datetime import datetime, timedelta, timezone
from time import sleep

import binance.enums as binance_enums
from binance.client import Client
from binance.enums import HistoricalKlinesType
from binance.exceptions import BinanceAPIException
from binance.helpers import convert_ts_str, interval_to_milliseconds
from django.apps import apps
from pytz import UTC
from requests.exceptions import ConnectionError, ConnectTimeout, ReadTimeout

from django_napse.utils.constants import DEFAULT_TAX, DOWNLOAD_STATUS, SIDES
from django_napse.utils.usefull_functions import round_down, round_up


class ExchangeController:
    pass


class BinanceController(ExchangeController):
    """Commmunication interface with binance account."""

    def __init__(self, public_key, private_key):
        client_created = False
        retries = 0
        while not client_created:
            try:
                self.client = Client(public_key, private_key)
                client_created = True
            except ConnectTimeout:
                print(f"ConnectionTimout error while creating client (try: {retries}), retrying...")
                time.sleep(1)
            except ConnectionError:
                print(f"ConnectionError error while creating client (try: {retries}), retrying...")
                time.sleep(1)
            except ReadTimeout:
                print(f"ReadTimeout error while creating client (try: {retries}), retrying...")
                time.sleep(1)
            else:
                if retries > 0:
                    print(f"Client created after {retries} retries.")
            retries += 1

        self.client.get_historical_klines = self._get_historical_klines
        self.client._historical_klines = self._historical_klines
        self.nb_retry: int = 3
        self.recv: int = 5000

    def get_info(self) -> dict:
        """Get the data of the binance account."""
        meter = 0
        while meter < self.nb_retry:
            meter += 1
            try:
                return self.client.get_account(recvWindow=self.recv)
            except ReadTimeout:
                continue
            except BinanceAPIException:
                continue

        return {"error": 408}

    def get_balance(self, ticker: str) -> dict:
        """Return balance of the selected ticker."""
        meter = 0
        while meter < self.nb_retry:
            meter += 1
            try:
                return self.client.get_asset_balance(asset=ticker.upper(), recvWindow=self.recv)
            except ReadTimeout:
                continue
            except BinanceAPIException:
                continue
        return {"error": 408}

    def get_tickers(self) -> list[dict]:
        """Return all available tickers."""
        meter = 0
        while meter < self.nb_retry:
            meter += 1
            try:
                return self.client.get_all_tickers()

            except ConnectionError:
                continue
        return [{"error": 408}]

    def buy_market(self, pair: str, quantity: float) -> dict:
        """Place a spot buy market order.

        !!! WARNING: quantity is in base currency !!!
        """
        Controller = apps.get_model("django_napse_core", "Controller")
        meter = 0
        while meter < self.nb_retry:
            meter += 1

            _, current_candle = Controller.get_candles(pair, "1m")
            price = current_candle["C"]
            quantity = quantity / price
            try:
                order = self.client.order_market_buy(symbol=pair.upper(), quantity=quantity, recvWindow=self.recv)
            except BinanceAPIException:
                continue
            else:
                return order
        return {"error": 408}

    def sell_market(self, pair: str, quantity: float) -> dict:
        """Place a sell market order.

        !!! WARNING: quantity is in quote currency !!!
        """
        meter = 0
        while meter < self.nb_retry:
            meter += 1

            try:
                order = self.client.order_market_sell(symbol=pair.upper(), quantity=quantity, recvWindow=self.recv)
            except BinanceAPIException:
                continue
            else:
                return order
        return {"error": 408}

    def test_market_order(self, pair: str, quantity: float, side_buy: bool = True) -> dict:
        """Simulate a binance order."""
        side = binance_enums.SIDE_BUY if side_buy else binance_enums.SIDE_SELL

        return self.client.create_test_order(
            symbol=pair.upper(),
            side=side,
            type=binance_enums.ORDER_TYPE_MARKET,
            quantity=quantity,
            # price = price,
            recvWindow=self.recv,
        )

    def get_historical_klines(
        self,
        dataset,
        start: datetime,
        end: datetime,
        pair: str,
        interval: str = "1m",
        limit: int = 1000,
        verbose: int = 0,
    ) -> None:
        start_str = str(start)
        end_str = str(end)

        self.client.get_historical_klines(
            pair=pair,
            interval=interval,
            start_str=start_str,
            end_str=end_str,
            limit=limit,
            dataset=dataset,
            verbose=verbose,
        )

    def _get_historical_klines(
        self,
        pair,
        interval,
        dataset,
        start_str,
        end_str,
        limit=1000,
        klines_type: HistoricalKlinesType = HistoricalKlinesType.SPOT,
        verbose: int = 0,
    ):
        dataset.completion = 0
        dataset.set_downloading()
        dataset.save()

        klines = self._historical_klines(
            pair=pair,
            interval=interval,
            start_str=start_str,
            end_str=end_str,
            limit=limit,
            klines_type=klines_type,
            dataset=dataset,
            verbose=verbose,
        )

        dataset.completion = 100
        dataset.set_idle()
        dataset.save()
        return klines

    def _historical_klines(
        self,
        pair,
        interval,
        dataset,
        start_str=None,
        end_str=None,
        limit=1000,
        klines_type: HistoricalKlinesType = HistoricalKlinesType.SPOT,
        verbose: int = 0,
    ):
        Candle = apps.get_model("django_napse_simulations", "Candle")

        output_data = []

        timeframe = interval_to_milliseconds(interval)

        start_ts = convert_ts_str(start_str)
        if start_ts is not None:
            first_valid_ts = self.client._get_earliest_valid_timestamp(pair, interval, klines_type)
            start_ts = max(start_ts, first_valid_ts)

        end_ts = convert_ts_str(end_str)
        total_loops = int(round_up((end_ts - start_ts) / limit / timeframe, 0))
        if end_ts and start_ts and end_ts <= start_ts:
            return output_data

        idx = 0
        start_time = time.time()
        last_percentage_saved = 0

        dataset.save()

        while True:
            found = False
            while not found:
                try:
                    temp_data = self.client._klines(
                        klines_type=klines_type,
                        symbol=pair,
                        interval=interval,
                        limit=limit,
                        startTime=start_ts,
                        endTime=end_ts,
                    )
                    found = True
                except ReadTimeout:
                    time.sleep(1)
                except ConnectionError:
                    time.sleep(1)
            dataset.create_candles(
                [
                    Candle(
                        dataset=dataset,
                        open_time=datetime.fromtimestamp(float(candle[0]) // 1000, tz=timezone.utc),
                        open=float(candle[1]),
                        high=float(candle[2]),
                        low=float(candle[3]),
                        close=float(candle[4]),
                        volume=float(candle[5]),
                    )
                    for candle in temp_data
                ],
            )

            idx += 1
            current_time = time.time()
            eta = (total_loops - idx) * (current_time - start_time) / idx
            eta = timedelta(seconds=eta)
            if verbose > 0:
                elapsed = timedelta(seconds=current_time - start_time)
                progress_str = f"Progress: {idx/total_loops*100:.2f} %"
                eta_str = f" (eta: {eta} s)"
                elapsed_str = f"  (elapsed: {elapsed})"
                loop_str = f" | Loop {idx}/{total_loops} completed"
                loop_size_str = f" | Loop size: {limit} candles."
                columns, rows = shutil.get_terminal_size()
                full_str = ""
                if columns <= len(progress_str):
                    pass
                elif columns <= len(progress_str + eta_str):
                    full_str = progress_str
                elif columns <= len(progress_str + eta_str + elapsed_str):
                    full_str = progress_str + eta_str
                elif columns <= len(progress_str + eta_str + elapsed_str + loop_str):
                    full_str = progress_str + eta_str + elapsed_str
                elif columns <= len(progress_str + eta_str + elapsed_str + loop_str + loop_size_str):
                    full_str = progress_str + eta_str + elapsed_str + loop_str
                else:
                    full_str = progress_str + eta_str + elapsed_str + loop_str + loop_size_str
                full_str += " " * (columns - len(full_str)) + "\r"
                print(full_str, end="")

            percentage = idx / total_loops * 100
            if percentage - last_percentage_saved > 0.1:
                last_percentage_saved = percentage
                dataset.completion = percentage
                dataset.last_update = datetime.now(tz=timezone.utc)
                dataset.eta = eta
                dataset.save()

            if not len(temp_data) or len(temp_data) < limit:
                break

            start_ts = temp_data[-1][0] + timeframe
            if end_ts and start_ts >= end_ts:
                break

        if verbose > 0:
            print()
        return output_data

    def fill_dataset(self, start_date, end_date, dataset, batch_size: int = 1000, verbose: int = 0):
        """Fill the dataset with historical data from the exchange."""
        if verbose > 1:
            print(
                f"### Starting to download: Dataset: pair={dataset.controller.pair}, interval={dataset.controller.interval}\n",
                f"\tstart={start_date}, end={end_date}",
                sep="",
            )
        self.get_historical_klines(
            start=start_date,
            end=end_date,
            pair=dataset.controller.pair,
            interval=dataset.controller.interval,
            limit=batch_size,
            dataset=dataset,
            verbose=verbose,
        )
        return dataset

    def download(
        self,
        controller,
        start_date: datetime,
        end_date: datetime,
        verbose: int = 0,
        *,
        squash: bool = False,
    ):
        """Download all the missing data to complete the dataset."""
        DataSet = apps.get_model("django_napse_simulations", "DataSet")

        start_time = datetime.now(tz=UTC)

        if end_date.second + end_date.microsecond != 0:
            end_date = end_date.replace(second=0, microsecond=0)
            end_date -= timedelta(microseconds=1)

        dataset = DataSet.objects.get(controller=controller)

        if datetime.now(tz=UTC) - dataset.last_update > timedelta(minutes=1):
            dataset.status = DOWNLOAD_STATUS.IDLE
            dataset.save()
        start_wait_time = time.time()
        while dataset.status == DOWNLOAD_STATUS.DOWNLOADING:
            time.sleep(0.1)
            if time.time() - start_wait_time > 10:
                error_msg = "Dataset is currently being downloaded. Come back later."
                raise TimeoutError(error_msg)

        if squash:
            dataset.delete()
            self.download(verbose=verbose)
        else:
            if dataset.start_date is None or dataset.end_date is None:
                dataset.save()
            if dataset.start_date is None or dataset.end_date is None:
                self.fill_dataset(start_date=start_date, end_date=end_date, dataset=dataset, verbose=verbose)
            else:
                if start_date < dataset.start_date:
                    end = end_date
                    end_date = dataset.start_date - timedelta(milliseconds=interval_to_milliseconds(dataset.controller.interval))
                    self.fill_dataset(start_date=start_date, end_date=end_date, dataset=dataset, verbose=verbose)
                    end_date = end
                if end_date > dataset.end_date:
                    start = start_date
                    start_date = dataset.end_date
                    self.fill_dataset(start_date=start_date, end_date=end_date, dataset=dataset, verbose=verbose)
                    start_date = start
        if verbose > 1:
            print(f"### Finished downloading at {datetime.now(tz=UTC)} (took {datetime.now(tz=UTC) - start_time} seconds)")
        dataset.save()
        return dataset

    def submit_order(
        self,
        controller,
        aggregated_order: dict,
        testing: bool,
    ) -> tuple[dict, dict, dict]:
        receipt = {}
        receipt[SIDES.BUY], executed_amounts_buy, fees_buy = self.send_order_to_exchange(
            side=SIDES.BUY,
            amount=aggregated_order["buy_amount"],
            controller=controller,
            min_trade=aggregated_order["min_trade"],
            price=aggregated_order["price"],
            testing=testing,
        )
        receipt[SIDES.SELL], executed_amounts_sell, feel_sell = self.send_order_to_exchange(
            side=SIDES.SELL,
            amount=aggregated_order["sell_amount"],
            controller=controller,
            min_trade=aggregated_order["min_trade"],
            price=aggregated_order["price"],
            testing=testing,
        )
        return receipt, executed_amounts_buy, executed_amounts_sell, fees_buy, feel_sell

    def send_order_to_exchange(
        self,
        controller,
        side: str,
        amount: float,
        min_trade: float,
        testing: bool,
        price: float,
    ) -> tuple[dict, dict]:
        if amount == 0:
            return {"error": "Amount too low"}, {}, {}

        executed_amounts = {}
        fees = {}
        if testing:
            if side == SIDES.BUY:
                amount /= price
                amount = round_down(amount, controller.lot_size)
                if amount > min_trade:
                    receipt = self.test_order(amount, SIDES.BUY, price, quote=controller.quote, base=controller.base)
                    exec_quote = -float(receipt["cummulativeQuoteQty"])
                    exec_base = 0
                    for elem in receipt["fills"]:
                        exec_base += float(elem["qty"]) - float(elem["commission"])
                        fees[elem["commissionAsset"]] = fees.get(elem["commissionAsset"], 0) + float(elem["commission"])
                    executed_amounts[controller.quote] = exec_quote
                    executed_amounts[controller.base] = exec_base

                else:
                    receipt = {"error": "Amount too low"}

            elif side == SIDES.SELL:
                amount = round_down(amount, controller.lot_size)
                if amount > min_trade:
                    receipt = self.test_order(amount, SIDES.SELL, price, quote=controller.quote, base=controller.base)
                    exec_quote = float(receipt["cummulativeQuoteQty"])
                    exec_base = -float(receipt["origQty"])
                    for elem in receipt["fills"]:
                        exec_quote -= float(elem["commission"])
                        fees[elem["commissionAsset"]] = fees.get(elem["commissionAsset"], 0) + float(elem["commission"])
                    executed_amounts[controller.quote] = exec_quote
                    executed_amounts[controller.base] = exec_base
                else:
                    receipt = {"error": "Amount too low"}

        else:
            # TODO: implement real orders
            error_msg = "IRL orders are not implemented yet. (failsafe to prevent accidental irl orders)."
            raise NotImplementedError(error_msg)
        return receipt, executed_amounts, fees

    def current_free_assets(self) -> dict:
        assets = self.get_info()["balances"]
        current = {}
        for elem in assets:
            if float(elem["free"]) > 0:
                current[elem.get("asset")] = float(elem.get("free"))
        return current

    @staticmethod
    def test_order(amount: float, side: str, price: float, base: str, quote: str) -> dict:
        if side not in (SIDES.BUY, SIDES.SELL):
            error_msg = f"Side must be BUY or SELL. Got {side}"
            raise ValueError(error_msg)

        pair = base + quote
        side_buy = side == SIDES.BUY

        executed_qty = amount
        cummulative_quote_qty = amount * price
        commission = executed_qty * DEFAULT_TAX["BINANCE"] / 100 if side_buy else cummulative_quote_qty * DEFAULT_TAX["BINANCE"] / 100

        commission_asset = base if side_buy else quote
        return {
            "symbol": pair,
            "orderId": None,
            "orderListId": None,
            "clientOrderId": None,
            "transactTime": None,
            "price": price,
            "origQty": amount,
            "executedQty": executed_qty,
            "cummulativeQuoteQty": cummulative_quote_qty,
            "status": "FILLED",
            "timeInForce": "GTC",
            "type": "MARKET",
            "side": side,
            "fills": [
                {
                    "price": price,
                    "qty": amount,
                    "commission": commission,
                    "commissionAsset": commission_asset,
                    "tradeId": None,
                },
            ],
        }

    def executed_amounts(self, receipt: dict, current_free_assets_dict: dict, controller, testing: bool) -> dict:
        executed = {}
        commission = 0
        if receipt == {} or receipt is None:
            return {}
        if testing:
            for fill in receipt.get("fills"):
                commission += float(fill.get("commission"))
            if receipt.get("side") == SIDES.BUY:
                base = receipt.get("fills")[0].get("commissionAsset")
                quote = receipt.get("symbol").replace(base, "")
                executed[base] = float(receipt.get("executedQty")) * (1 - DEFAULT_TAX["BINANCE"] / 100)
                executed[quote] = -float(receipt.get("cummulativeQuoteQty"))
            elif receipt.get("side") == SIDES.SELL:
                quote = receipt.get("fills")[0].get("commissionAsset")
                base = receipt.get("symbol").replace(quote, "")
                executed[base] = -float(receipt.get("executedQty"))
                executed[quote] = float(receipt.get("cummulativeQuoteQty")) * (1 - DEFAULT_TAX["BINANCE"] / 100)
        else:
            while self.current_free_assets(controller) == current_free_assets_dict:
                print("Waiting for order to be executed...")
                sleep(0.01)
            new_free_assets = self.current_free_assets(controller)
            for asset, amount in new_free_assets.items():
                if amount != current_free_assets_dict.get(asset):
                    executed[asset] = amount - current_free_assets_dict.get(asset)
        return executed



from django.db import models

from django_napse.core.models.bots.managers.plugin import PluginManager
from django_napse.utils.findable_class import FindableClass


class Plugin(models.Model, FindableClass):
    """A plugin is a pluggable extention to a bot."""

    strategy = models.ForeignKey("Strategy", on_delete=models.CASCADE, related_name="plugins")
    category = models.CharField(max_length=255)

    objects = PluginManager()

    def __str__(self) -> str:
        return f"PLUGIN {self.pk=}"

    def info(self, beacon: str = "", *, verbose: bool = True) -> str:
        """Return a string with the model information.

        Args:
            beacon (str, optional): The prefix for each line. Defaults to "".
            verbose (bool, optional): Whether to print the string. Defaults to True.

        Returns:
            str: The string with the history information.
        """
        string = ""
        string += f"{beacon}Plugin {self.pk}:\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.category=}\n"

        if verbose:
            print(string)
        return string

    @classmethod
    def plugin_category(cls):  # noqa
        if cls == Plugin:
            error_msg = "plugin_category not implemented for the Plugin base class, please implement it in a subclass."
        else:
            error_msg = f"plugin_category not implemented for the Plugin base class, please implement it in the {cls} class."
        raise NotImplementedError(error_msg)

    def _connect(self, connection):  # noqa
        if self.__class__ == Plugin:
            error_msg = "connect not implemented for the Plugin base class, please implement it in a subclass."
        else:
            error_msg = f"connect not implemented for the Plugin base class, please implement it in the {self.__class__} class."
        raise NotImplementedError(error_msg)

    def connect(self, connection):  # noqa
        self = self.find()
        return self._connect(connection)

    def apply__no_db(self, data: dict) -> dict:
        if self.__class__ == Plugin:
            error_msg = "apply__no_db not implemented for the Plugin base class, please implement it in a subclass."
        else:
            error_msg = f"apply__no_db not implemented for the Plugin base class, please implement it in the {self.__class__} class."
        raise NotImplementedError(error_msg)

    def apply(self, data: dict) -> dict:
        self = self.find()
        return self.apply__no_db(data)



from datetime import timedelta

import pandas as pd
from binance.helpers import interval_to_milliseconds
from django.db import models

from django_napse.simulations.models.datasets.managers.dataset import DataSetManager
from django_napse.utils.constants import DOWNLOAD_STATUS
from django_napse.utils.errors import DataSetError


class DataSet(models.Model):
    controller = models.OneToOneField("django_napse_core.Controller", on_delete=models.CASCADE, related_name="dataset")
    start_date = models.DateTimeField(null=True, blank=True)
    end_date = models.DateTimeField(null=True, blank=True)

    last_update = models.DateTimeField(auto_now_add=True)
    status = models.CharField(max_length=12, default=DOWNLOAD_STATUS.IDLE)
    completion = models.FloatField(default=0.0)
    eta = models.DurationField(null=True, blank=True)

    objects = DataSetManager()

    def __str__(self):
        return f"DATASET: {self.pk=}"

    def save(self, *args, **kwargs):
        if self.completion < 0 or self.completion > 100:
            error_msg = f"Completion ({self.completion}) not in [0, 100]"
            raise DataSetError.InvalidSettings(error_msg)

        if self.status not in str(DOWNLOAD_STATUS):
            error_msg = f"Status ({self.status}) not in {DOWNLOAD_STATUS}"
            raise DataSetError.InvalidSettings(error_msg)

        query = Candle.objects.filter(dataset=self).order_by("open_time")
        number_of_candles = query.count()
        if number_of_candles > 0:
            self.start_date = query[0].open_time
            self.end_date = query[number_of_candles - 1].open_time + timedelta(milliseconds=interval_to_milliseconds(self.controller.interval) - 1)
        super().save(*args, **kwargs)

    def info(self, verbose=True, beacon=""):
        string = ""
        string += f"{beacon}Dataset {self.pk}:\n"
        string += f"{beacon}\t{self.controller=}\n"
        string += f"{beacon}\t{self.start_date=}\n"
        string += f"{beacon}\t{self.end_date=}\n"
        string += f"{beacon}\t{self.last_update=}\n"
        string += f"{beacon}\t{self.status=}\n"
        string += f"{beacon}\t{self.completion=}\n"
        string += f"{beacon}\t{self.eta=}\n"
        string += f"{beacon}Candles:\n"
        candles = self.candles.all().order_by("open_time")
        if candles.count() > 10:
            for candle in candles[:5]:
                string += f"{beacon}\tO:{candle.open_time}\t H: {candle.high}\t L: {candle.close}\t C: {candle.low}\t V: {candle.volume}\n"
            string += f"{beacon}\t...\n"
            for candle in candles[candles.count() - 5 :]:
                string += f"{beacon}\tO:{candle.open_time}\t H: {candle.high}\t L: {candle.close}\t C: {candle.low}\t V: {candle.volume}\n"
            string += f"{beacon}\t({candles.count()} candles)\n"
        else:
            for candle in candles:
                string += f"{beacon}\tO:{candle.open_time}\t H: {candle.high}\t L: {candle.close}\t C: {candle.low}\t V: {candle.volume}\n"

        if verbose:
            print(string)
        return string

    def create_candles(self, candles: list):
        """Save a list of candle Objects into the database."""
        Candle.objects.bulk_create(candles)

    def set_downloading(self):
        """Set the dataset status to downloading.

        Raises:
            ValueError: If the dataset isn't in IDLE status.
        """
        if self.status == DOWNLOAD_STATUS.IDLE:
            self.status = DOWNLOAD_STATUS.DOWNLOADING
            self.completion = 0
            self.save()
        else:
            error_msg = "Dataset is already downloading."
            raise DataSetError.InvalidSettings(error_msg)

    def set_idle(self):
        """Set the dataset status to idle."""
        if self.status == DOWNLOAD_STATUS.DOWNLOADING:
            self.status = DOWNLOAD_STATUS.IDLE
            self.save()
        else:
            error_msg = "Dataset is not downloading."
            raise DataSetError.InvalidSettings(error_msg)

    def is_finished(self):
        return self.status == DOWNLOAD_STATUS.IDLE and self.completion == 100

    def to_dataframe(self, start_date=None, end_date=None):
        start_date = start_date or self.start_date
        end_date = end_date or self.end_date
        candles = self.candles.filter(open_time__gte=start_date, open_time__lt=end_date).order_by("open_time")
        return pd.DataFrame(
            data={
                "open_time": [candle.open_time for candle in candles],
                "open": [candle.open for candle in candles],
                "high": [candle.high for candle in candles],
                "low": [candle.low for candle in candles],
                "close": [candle.close for candle in candles],
                "volume": [candle.volume for candle in candles],
            },
        )


class Candle(models.Model):
    dataset = models.ForeignKey("DataSet", on_delete=models.CASCADE, related_name="candles")
    open_time = models.DateTimeField()
    open = models.FloatField()
    high = models.FloatField()
    low = models.FloatField()
    close = models.FloatField()
    volume = models.FloatField()

    class Meta:
        unique_together = ("dataset", "open_time")

    def __str__(self):
        return f"CANDLE: {self.pk=}"

    def info(self, verbose=True, beacon=""):
        string = ""
        string += f"{beacon}Candle {self.pk}:\n"
        string += f"{beacon}\t{self.dataset=}\n"
        string += f"{beacon}\t{self.open_time=}\n"
        string += f"{beacon}\t{self.open=}\n"
        string += f"{beacon}\t{self.high=}\n"
        string += f"{beacon}\t{self.low=}\n"
        string += f"{beacon}\t{self.close=}\n"
        string += f"{beacon}\t{self.volume=}\n"

        if verbose:
            print(string)
        return string

    def to_dict(self):
        return {
            "controller": self.dataset.controller,
            "open_time": self.open_time,
            "open": self.open,
            "high": self.high,
            "low": self.low,
            "close": self.close,
            "volume": self.volume,
            "extra": {},
        }


class DataSetQueue(models.Model):
    controller = models.ForeignKey("django_napse_core.Controller", on_delete=models.CASCADE, related_name="dataset_queues")
    start_date = models.DateTimeField()
    end_date = models.DateTimeField()

    created_at = models.DateTimeField(auto_now_add=True)

    def __str__(self) -> str:
        return f"DATASET_QUEUE: {self.pk=}"

    def get_dataset(self) -> DataSet:
        try:
            return DataSet.objects.get(controller=self.controller)
        except DataSet.DoesNotExist:
            return None
            return None



import requests
from django.core.exceptions import ValidationError

from django_napse.core.models.bots.controller import Controller
from django_napse.core.tasks.base_task import BaseTask


class CandleCollectorTask(BaseTask):
    """Task to collect candles from binance's api and send it to controllers."""

    name = "candle_collector"
    interval_time = 30

    @staticmethod
    def build_candle(request_json: list[list[int], list[int]]) -> tuple[dict[str, int | float], dict[str, int | float]]:
        """Structure close_candle & current candle from the request.json().

        Candle shape: {"T": 1623000000000, "O": 1.0, "H": 1.0, "L": 1.0, "C": 1.0, "V": 1.0}.

        Args:
        ----
        request_json: response of binance's api

        Returns:
        -------
            closed_candle (dict[str, int | float]): last closed candle
            current_candle (dict[str, int | float]): current candle

        Raises:
        ------
            ValidationError: if request_json is not a list of 2 lists
        """
        dico_structure_label = ["T", "O", "H", "L", "C", "V"]

        # Check request
        if not isinstance(request_json, list):
            error_msg = f"request_json is not a list ({type(request_json)})"
            raise ValidationError(error_msg)

        if len(request_json) != 2:  # noqa
            error_msg = f"request_json must have 2 elements ({len(request_json)})"
            raise ValidationError(error_msg)

        if not isinstance(request_json[0], list) or not isinstance(request_json[1], list):
            error_msg = f"request_json is not a list of 2 lists ({type(request_json[0])}, {type(request_json[1])})"
            raise ValidationError(error_msg)

        if len(request_json[0]) != len(request_json[1]):
            error_msg = f"The 2 lists of requests json must have the same length ({len(request_json[0])}, {len(request_json[1])})"
            raise ValidationError(error_msg)

        if len(request_json[0]) < len(dico_structure_label) or len(request_json[1]) < len(dico_structure_label):
            error_msg = f"Request_json's lists are too short ({len(request_json[0])}, {len(request_json[1])})"
            raise ValidationError(error_msg)

        # Build candle with the good shape
        closed_candle: dict[str, int | float] = {}
        current_candle: dict[str, int | float] = {}
        for i, label in enumerate(dico_structure_label):
            if isinstance(request_json[0][i], int):
                closed_candle[label] = request_json[0][i]
                current_candle[label] = request_json[1][i]
            else:
                closed_candle[label] = float(request_json[0][i])
                current_candle[label] = float(request_json[1][i])
        return closed_candle, current_candle

    @staticmethod
    def request_get(pair: str, interval: str, api: str = "api") -> requests.Response:
        """Make a request to binance's api to get 2 candles (the last closed and the current one).

        Args:
        ----
        pair (str): pair of the candles
        interval (str): interval of the candles
        api (str, optional): api to use

        Returns:
        -------
            response of the request (requests.Response)
        """
        url = f"https://{api}.binance.com/api/v3/klines?symbol={pair}&interval={interval}&limit=2"
        try:
            req = requests.get(url, timeout=10)
        except requests.exceptions.ConnectionError:
            req = None
        return req

    def get_candles(self, pair: str, interval: str) -> tuple[dict[str, int | float], dict[str, int | float]]:
        """Get candles from binance's api.

        Retry automatically on all binance's backup api if the request failed.

        Args:
        ----
        pair: pair of the candles
        interval: interval of the candles

        Returns:
        -------
            closed_candle, current_candle

        Raises:
        ------
            ValidationError: if request failed on all apis
        """
        apis = ("api", "api1", "api2", "api3")

        for api in apis:
            request = self.request_get(pair, interval, api=api)
            # Valid request
            success_code = 200
            if request.status_code == success_code:
                return self.build_candle(request.json())
        # All requests failed
        error_msg = f"Impossible to get candles from binance's api (pair: {pair}, interval: {interval})"
        raise ValueError(error_msg)

    def run(self) -> None:
        """Run the task.

        Try to get the results of request of binance's api and send it to controller(s).
        If the request failed, the controller(s) is add to a list and controller(s) is this list try again (on all binance's backup api) at the end.
        """
        if not self.avoid_overlap(verbose=False):
            return
        self.info("Running CandleCollectorTask")

        failed_controllers: list["Controller"] = []
        failed_controllers_second_attempt: list["Controller"] = []
        all_orders = []
        for controller in Controller.objects.all():
            request = self.request_get(controller.pair, controller.interval, "api")
            success_code = 200
            if request is None or request.status_code != success_code:
                warning = f"Controller {controller.pk} failed on 'api'"
                self.warning(warning)
                failed_controllers.append(controller)
                continue
            closed_candle, current_candle = self.build_candle(request.json())
            all_orders += controller.send_candles_to_bots(closed_candle, current_candle)

        backup_apis = ("api1", "api2", "api3")
        for controller in failed_controllers:
            success = False
            for api in backup_apis:
                request = self.request_get(controller.pair, controller.interval, api)
                success_code = 200
                if request is not None and request.status_code == success_code:
                    closed_candle, current_candle = self.build_candle(request.json())
                    info = f"{controller} succeeded on '{api}'"
                    self.info(info)
                    success = True
                    break
            if success:
                all_orders += controller.send_candles_to_bots(closed_candle, current_candle)
            else:
                failed_controllers_second_attempt.append(controller)

        for controller in failed_controllers_second_attempt:
            error = f"{controller} failed on all apis"
            self.error(error)


CandleCollectorTask().delete_task()
CandleCollectorTask().register_task()



from django_napse.core.models.bots.architectures.single_pair import SinglePairArchitecture
from django_napse.core.models.bots.implementations.empty.config import EmptyBotConfig
from django_napse.core.models.bots.strategy import Strategy
from django_napse.utils.constants import SIDES


class EmptyStrategy(Strategy):
    def __str__(self) -> str:
        return f"EMPTY BOT STRATEGY: {self.pk=}"

    def info(self, verbose=True, beacon=""):
        string = ""
        string += f"{beacon}Strategy ({self.pk=}):\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.config=}\n"
        string += f"{beacon}\t{self.architecture=}\n"
        if verbose:  # pragma: no cover
            print(string)
        return string

    @classmethod
    def config_class(cls):
        return EmptyBotConfig

    @classmethod
    def architecture_class(cls):
        return SinglePairArchitecture

    def give_order(self, data: dict) -> list[dict]:
        controller = data["controllers"]["main"]
        return [
            {
                "controller": controller,
                "ArchitectureModifications": [],
                "StrategyModifications": [],
                "ConnectionModifications": [],
                "connection": data["connection"],
                "asked_for_amount": 0,
                "asked_for_ticker": controller.quote,
                "pair": controller.pair,
                "price": data["candles"][controller]["latest"]["close"],
                "side": SIDES.KEEP,
            },
        ]



from typing import ClassVar

from rest_framework import serializers

from django_napse.auth.models import KeyPermission


class PermissionSerializer(serializers.ModelSerializer):
    """Serialize a permission instance."""

    space = serializers.CharField(source="space.uuid")

    class Meta:  # noqa: D106
        model = KeyPermission
        fields: ClassVar[list[str]] = [
            "uuid",
            "permission_type",
            "approved",
            "space",
        ]



from datetime import datetime

from django.db import models


class DataSetManager(models.Manager):
    def create(self, controller, start_date: datetime, end_date: datetime):
        try:
            dataset = self.model.objects.get(controller=controller)
        except self.model.DoesNotExist:
            dataset = self.model(
                controller=controller,
            )
            dataset.save()
        return dataset.controller.download(
            start_date=start_date,
            end_date=end_date,
        )



from typing import TYPE_CHECKING

from django.db import models

from django_napse.core.models.connections.managers import ConnectionManager
from django_napse.core.models.transactions.transaction import Transaction
from django_napse.utils.constants import TRANSACTION_TYPES
from django_napse.utils.usefull_functions import process_value_from_type

if TYPE_CHECKING:
    from django_napse.core.models.accounts.space import Space


class Connection(models.Model):
    """Link between a bot & a wallet."""

    owner = models.ForeignKey("Wallet", on_delete=models.CASCADE, related_name="connections")
    bot = models.ForeignKey("Bot", on_delete=models.CASCADE, related_name="connections")

    created_at = models.DateTimeField(auto_now_add=True, blank=True)

    objects = ConnectionManager()

    class Meta:  # noqa: D106
        unique_together = ("owner", "bot")

    def __str__(self) -> str:  # pragma: no cover
        return f"CONNECTION: {self.pk=}"

    def info(self, verbose: bool = True, beacon: str = "") -> str:  # noqa: FBT001, FBT002
        """Give info on the Connection instance."""
        string = ""
        string += f"{beacon}Connection {self.pk}:\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.owner=}\n"
        string += f"{beacon}\t{self.bot=}\n"
        string += f"{beacon}\t{self.created_at=}\n"

        string += f"{beacon}Wallet:\n"
        new_beacon = beacon + "\t"
        string += f"{self.wallet.info(verbose=False, beacon=new_beacon)}\n"

        string += f"{beacon}ConnectionSpecificArgs:\n"
        query = self.specific_args.all()
        if query.count() == 0:
            string += f"{beacon}\tNone\n"
        else:
            for connection_specific_arg in query:
                string += f"{connection_specific_arg.info(verbose=False, beacon=new_beacon)}\n"

        if verbose:  # pragma: no cover
            print(string)
        return string

    @property
    def testing(self) -> bool:
        """Return true if the connection's space is in testing."""
        return self.space.testing

    @property
    def space(self) -> "Space":
        """Return the space of the connection."""
        return self.owner.find().space

    def deposit(self, ticker: str, amount: int) -> Transaction:
        """Make a transaction from owner's wallet to bot's wallet."""
        return Transaction.objects.create(
            from_wallet=self.owner,
            to_wallet=self.wallet,
            amount=amount,
            ticker=ticker,
            transaction_type=TRANSACTION_TYPES.CONNECTION_DEPOSIT,
        )

    def withdraw(self, ticker: str, amount: int) -> Transaction:
        """Make a transaction from bot's wallet to owner's wallet."""
        return Transaction.objects.create(
            from_wallet=self.wallet,
            to_wallet=self.owner,
            amount=amount,
            ticker=ticker,
            transaction_type=TRANSACTION_TYPES.CONNECTION_WITHDRAW,
        )

    def to_dict(self) -> dict[str, any]:
        """Return a connection to a dictionary."""
        return {
            "connection": self,
            "wallet": self.wallet.find().to_dict(),
            "connection_specific_args": {arg.key: arg for arg in self.specific_args.all()},
        }


class ConnectionSpecificArgs(models.Model):
    """Argument of a connection.

    A connection can have a infinite number of arguments.
    """

    connection = models.ForeignKey("Connection", on_delete=models.CASCADE, related_name="specific_args")
    key = models.CharField(max_length=100)
    value = models.CharField(max_length=100, default="None")
    target_type = models.CharField(max_length=100, default="None")

    class Meta:  # noqa: D106
        unique_together = ("connection", "key")

    def __str__(self) -> str:  # pragma: no cover
        string = f"CONNECTION_SPECIFIC_ARGS: {self.pk=},"
        return string + f"connection__pk={self.connection.pk}, key={self.key}, value={self.value}, target_type={self.target_type}"

    def info(self, verbose: bool = True, beacon: str = "") -> str:  # noqa: FBT001, FBT002
        """Give info on the ConnectionSpecificArgs instance."""
        string = ""
        string += f"{beacon}ConnectionSpecificArgs {self.pk}:\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.connection=}\n"
        string += f"{beacon}\t{self.key=}\n"
        string += f"{beacon}\t{self.value=}\n"
        string += f"{beacon}\t{self.target_type=}\n"
        if verbose:  # pragma: no cover
            print(string)
        return string

    def get_value(self) -> any:
        """Return the value in the correct type."""
        return process_value_from_type(self.value, self.target_type)



from django.db import models

# from django_napse.core.models.connections.connection import Connection
from django_napse.core.models.fleets.link import Link
from django_napse.core.models.transactions.transaction import Transaction
from django_napse.utils.constants import TRANSACTION_TYPES
from django_napse.utils.errors import BotError, ClusterError


class Cluster(models.Model):
    fleet = models.ForeignKey(
        "Fleet",
        on_delete=models.CASCADE,
        related_name="clusters",
    )
    template_bot = models.OneToOneField(
        "Bot",
        on_delete=models.CASCADE,
        related_name="cluster",
    )
    share = models.FloatField()
    breakpoint = models.FloatField()
    autoscale = models.BooleanField()

    def __str__(self):
        return f"Cluster: {self.fleet}"

    def save(self, *args, **kwargs):
        if not self.config.immutable:
            error_msg = "In a fleet, the config must be immutable."
            raise ClusterError.MutableBotConfig(error_msg)
        return super().save(*args, **kwargs)

    def info(self, verbose=True, beacon=""):
        string = ""
        string += f"{beacon}Cluster {self.pk}:\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.fleet=}\n"
        string += f"{beacon}\t{self.template_bot=}\n"
        string += f"{beacon}\t{self.share=}\n"
        string += f"{beacon}\t{self.breakpoint=}\n"
        string += f"{beacon}\t{self.autoscale=}\n"

        new_beacon = beacon + "\t"
        string += f"{beacon}Links:\n"
        for link in self.links.all():
            string += f"{link.info(verbose=False, beacon=new_beacon)}\n"

        string += f"{beacon}Connections:\n"
        connections = []
        for link in self.links.all():
            connections += list(link.bot.connections.all())
        for connection in connections:
            string += f"{connection.info(verbose=False, beacon=new_beacon)}\n"

        if verbose:  # pragma: no cover
            print(string)
        return string

    @property
    def config(self):
        return self.template_bot.strategy.config

    def invest(self, space, amount, ticker):
        all_connections = []
        bots = [link.bot for link in self.links.all().order_by("importance")]
        if len(bots) == 0:
            new_bot = self.template_bot.copy()
            Link.objects.create(bot=new_bot, cluster=self, importance=1)
            # connection = Connection.objects.create(bot=new_bot, owner=space.wallet)
            connection = space.wallet.connect_to_bot(new_bot)
            Transaction.objects.create(
                from_wallet=space.wallet,
                to_wallet=connection.wallet,
                amount=amount,
                ticker=ticker,
                transaction_type=TRANSACTION_TYPES.CONNECTION_DEPOSIT,
            )
            all_connections.append(connection)
        elif len(bots) == 1:
            bot = bots[0]
            if ticker not in bot.architecture.accepted_investment_tickers() or ticker not in bot.architecture.accepted_tickers():
                error_msg = f"Bot {bot} does not accept ticker {ticker}."
                raise BotError.InvalidTicker(error_msg)

            connection = space.wallet.connect_to_bot(bot)
            # connection = Connection.objects.get(bot=bot, owner=space.wallet)
            Transaction.objects.create(
                from_wallet=space.wallet,
                to_wallet=connection.wallet,
                amount=amount,
                ticker=ticker,
                transaction_type=TRANSACTION_TYPES.CONNECTION_DEPOSIT,
            )
            all_connections.append(connection)
        else:
            error_msg = "Autoscale not implemented yet."
            raise NotImplementedError(error_msg)
        return all_connections



from django.db import models

from django_napse.utils.constants import MODIFICATION_STATUS
from django_napse.utils.usefull_functions import process_value_from_type

from .modification import Modification


class ArchitectureModification(Modification):
    def apply(self):
        architecture = self.order.connection.bot.architecture.find()
        architectur, self = self.apply__no_db(architecture)
        architecture.save()
        self.save()

    def apply__no_db(self, **kwargs: dict) -> tuple[models.Model, "ArchitectureModification"]:
        architecture = kwargs.get("architecture")
        if not hasattr(architecture, f"variable_{self.key}"):
            error_msg: str = f"Architecture {architecture} must have attribute {self.key}"
            raise ValueError(error_msg)

        setattr(architecture, f"variable_{self.key}", process_value_from_type(self.value, self.target_type))
        self.status = MODIFICATION_STATUS.APPLIED
        return architecture, self



from django.urls import reverse
from rest_framework.test import APIClient
from rest_framework_api_key.permissions import HasAPIKey

from django_napse.api.custom_permissions import HasAdminPermission, HasFullAccessPermission, HasMasterKey, HasReadPermission, HasSpace
from django_napse.auth.models import KeyPermission
from django_napse.auth.models.keys.key import NapseAPIKey
from django_napse.utils.constants import PERMISSION_TYPES
from django_napse.utils.custom_test_case import CustomTestCase


class APITestCase(CustomTestCase):
    def setUp(self):
        self.client = APIClient()
        self.modes = self.setup()
        if self.modes == {} or not isinstance(self.modes, dict):
            error_msg = "Modes not defined"
            raise ValueError(error_msg)

    def setup(self):
        error_msg = "Setup function not implemented"
        raise NotImplementedError(error_msg)

    def build_url(self, kwargs=None):
        if kwargs and kwargs.get("pk", None):
            self.url = reverse(viewname=self.url_name, kwargs={"pk": kwargs.pop("pk")})
        else:
            self.url = reverse(viewname=self.url_name)
        if kwargs:
            self.url += "?" + "&".join([f"{key}={value}" for key, value in kwargs.items()])

    def build_key(self, permissions):
        key_obj, self.key = NapseAPIKey.objects.create_key(name="test")
        if HasReadPermission in permissions:
            KeyPermission.objects.create(key=key_obj, space=self.space, permission_type=PERMISSION_TYPES.READ_ONLY, approved=True)
        if HasFullAccessPermission in permissions:
            KeyPermission.objects.create(key=key_obj, space=self.space, permission_type=PERMISSION_TYPES.FULL_ACCESS, approved=True)
        if HasAdminPermission in permissions:
            KeyPermission.objects.create(key=key_obj, space=self.space, permission_type=PERMISSION_TYPES.ADMIN, approved=True)

    def authenticate(self):
        self.headers["Authorization"] = f"Api-Key {self.key}"

    def logout(self):
        self.headers.pop("Authorization")

    def request(self, method):
        match method:
            case "GET":
                return self.client.get(path=self.url, data=self.data, headers=self.headers)
            case "POST":
                return self.client.post(path=self.url, data=self.data, headers=self.headers)
            case "PUT":
                return self.client.put(path=self.url, data=self.data, headers=self.headers)
            case "PATCH":
                return self.client.patch(path=self.url, data=self.data, headers=self.headers)
            case "DELETE":
                return self.client.delete(path=self.url, data=self.data, headers=self.headers)
            case _:
                error_msg = f"Unknown method: {method}"
                raise ValueError(error_msg)

    def check_auth(self, name: str, mode: str, error_list: list, divider: int = 1, expected: int = 403) -> None:
        response = self.request(self.modes[mode]["method"])
        try:
            self.assertEqual(response.status_code // divider, expected)
        except AssertionError as e:
            e.add_note(f"{name} test failed (no permissions didn't return 403)")
            e.add_note(str(response.data))
            error_list.append(e)

    def run_tests(self, mode):
        if mode not in self.modes:
            error_msg = f"Unknown mode: {mode}"
            raise ValueError(error_msg)

        error_list = []
        self.kwargs = self.modes[mode].get("kwargs", {})
        self.data = self.modes[mode].get("data", {})
        self.headers = self.modes[mode].get("headers", {})
        self.url_name = self.modes[mode]["url_name"]
        self.build_url(kwargs={"pk": 1} if self.modes[mode]["method"] in ["PATCH", "PUT", "DELETE"] and not self.kwargs else self.kwargs)

        permissions = self.modes[mode]["view"].permission_classes
        permission_importance = {
            HasSpace: 0,
            HasAPIKey: 1,
            HasReadPermission: 2,
            HasFullAccessPermission: 3,
            HasAdminPermission: 4,
            HasMasterKey: 5,
        }
        permissions.sort(key=lambda x: permission_importance[x])

        if permissions == []:
            self.check_auth(name="No permissions", mode=mode, error_list=error_list, divider=100, expected=2)
        else:
            if HasSpace in permissions:
                self.check_auth(name="HasSpace with no key", mode=mode, error_list=error_list, expected=400)
                self.build_url(kwargs={"space": str("random uuid"), **self.kwargs})
                self.check_auth(name="HasSpace with no key", mode=mode, error_list=error_list, expected=400)
                self.build_url(kwargs={"space": str("7aafc68d-f619-4874-aaf5-c123a176e303"), **self.kwargs})
                self.check_auth(name="HasSpace with no key", mode=mode, error_list=error_list, expected=400)
                self.build_url(kwargs={"space": str(self.space.uuid), **self.kwargs})

            if HasAPIKey in permissions:
                self.check_auth(name="HasAPIKey with no key", mode=mode, error_list=error_list)

            if HasReadPermission in permissions:
                self.check_auth(name="HasReadPermission with no key", mode=mode, error_list=error_list)
                self.build_key([])
                self.authenticate()
                self.check_auth(name="HasReadPermission with no permissions", mode=mode, error_list=error_list)

            if HasFullAccessPermission in permissions:
                self.check_auth(name="HasFullAccessPermission with no key", mode=mode, error_list=error_list)
                self.build_key([])
                self.authenticate()
                self.check_auth(name="HasFullAccessPermission with no permissions", mode=mode, error_list=error_list)
                self.build_key([HasReadPermission])
                self.authenticate()
                self.check_auth(name="HasFullAccessPermission with read permissions", mode=mode, error_list=error_list)

            if HasAdminPermission in permissions:
                self.check_auth(name="HasAdminPermission with no key", mode=mode, error_list=error_list)
                self.build_key([])
                self.authenticate()
                self.check_auth(name="HasAdminPermission with no permissions", mode=mode, error_list=error_list)
                self.build_key([HasReadPermission])
                self.authenticate()
                self.check_auth(name="HasAdminPermission with read permissions", mode=mode, error_list=error_list)
                self.build_key([HasReadPermission, HasFullAccessPermission])
                self.authenticate()
                self.check_auth(name="HasAdminPermission with read and full access permissions", mode=mode, error_list=error_list)

        self.build_key(permissions)
        self.authenticate()

        for test in self.modes[mode]["tests"]:
            response = test["setup"]()
            try:
                self.assertEqual(test["status_code"], response.status_code)
            except AssertionError as e:
                e.add_note(f"Test: {test['name']} | {test['setup'].__name__} | status code")
                e.add_note(str(response.data))
                error_list.append(e)

            for check in test["checks"]:
                try:
                    self.assertTrue(check(response))
                except Exception as e:
                    e.add_note(f"Test: {test['name']} | {test['setup'].__name__} | check {check.__name__}")
                    e.add_note(str(response.data))
                    error_list.append(e)

        if len(error_list) > 0:
            error_msg = f"Errors in {mode} mode"
            raise ExceptionGroup(error_msg, error_list)


class ViewTest:
    def __init__(self, testcase_instance: APITestCase, *args, **kwargs):
        self.testcase_instance = testcase_instance

    def setup(self, data: dict | None = None):
        def _setup(data=data):
            # TODO: Add other methods (post, put, patch, delete)
            return self.testcase_instance.client.get(
                path=self.testcase_instance.url,
                data=data,
                headers=self.testcase_instance.headers,
            )

        return _setup

    def run(self) -> list[dict[str, any]]:
        raise NotImplementedError



from django_napse.core.models import Controller
from django_napse.core.tasks.base_task import BaseTask


class ControllerUpdateTask(BaseTask):
    """Task to update all controllers."""

    name = "controller_update"
    interval_time = 45
    time_limit = 60
    soft_time_limit = 60

    def run(self) -> None:
        """Run a task to update all controllers."""
        if not self.avoid_overlap(verbose=False):
            return
        self.info("Running ControllerUpdateTask")
        for controller in Controller.objects.all():
            controller.update_variables_always()
            controller.get_price_always()


ControllerUpdateTask().delete_task()
ControllerUpdateTask().register_task()



from typing import ClassVar

from django_napse.core.models import Bot, Controller, EmptyStrategy
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.bots.test_bot -v2 --keepdb --parallel
"""


class BotDefaultTestCase:
    model = Bot
    strategy_class = EmptyStrategy
    config_settings: ClassVar = {"empty": True}

    @property
    def architecture_constants(self):
        return {
            "controller": Controller.get(
                exchange_account=self.exchange_account,
                base="BTC",
                quote="USDT",
                interval="1d",
            ),
        }

    def simple_create(self):
        return self.model.objects.create(name="Test Bot", strategy=self.strategy)

    @property
    def config(self):
        return self.strategy_class.config_class().objects.create(space=self.space, settings=self.config_settings)

    @property
    def architecture(self):
        return self.strategy_class.architecture_class().objects.create(constants=self.architecture_constants)

    @property
    def strategy(self):
        return self.strategy_class.objects.create(config=self.config, architecture=self.architecture)


class BotBINANCETestCase(BotDefaultTestCase, ModelTestCase):
    exchange = "BINANCE"



from django_napse.api.fleets import FleetView
from django_napse.core.models import Bot, Controller, EmptyBotConfig, EmptyStrategy, Fleet, SinglePairArchitecture
from django_napse.utils.api_test_case import APITestCase, ViewTest
from django_napse.utils.dict_comparison import compare_responses

"""
python tests/test_app/manage.py test tests.django_tests.api.fleets.test_fleet_view -v2 --keepdb --parallel
"""


class ListFleetViewTest(ViewTest):
    def __init__(self, *args, **kwargs):
        """Build a fleet."""
        super().__init__(*args, **kwargs)
        # create simple fleet
        config = EmptyBotConfig.objects.create(space=self.testcase_instance.space, settings={"empty": True})
        # config = EmptyBotConfig.objects.get(space=self.testcase_instance.space)
        print(f"config: {config}")
        controller = Controller.get(
            exchange_account=self.testcase_instance.exchange_account,
            base="BTC",
            quote="USDT",
            interval="1m",
        )
        architecture = SinglePairArchitecture.objects.create(constants={"controller": controller})
        strategy = EmptyStrategy.objects.create(config=config, architecture=architecture)
        bot = Bot.objects.create(name="Test Bot", strategy=strategy)
        self.fleet = Fleet.objects.create(
            name="Test Fleet",
            exchange_account=self.testcase_instance.exchange_account,
            clusters=[
                {
                    "template_bot": bot,
                    "share": 0.7,
                    "breakpoint": 1000,
                    "autoscale": False,
                },
                {
                    "template_bot": bot,
                    "share": 0.3,
                    "breakpoint": 1000,
                    "autoscale": True,
                },
            ],
        )

    def check_response(self, response) -> bool:
        return compare_responses(
            response.data[0],
            {
                "uuid": str(),
                "name": str(),
                "value": float(),
                "bot_count": int(),
            },
        )

    def run(self):
        print("list run")
        return [
            {
                "name": "response",
                "setup": self.setup(),
                "status_code": 200,
                "checks": [self.check_response],
            },
        ]


class RetrieveFleetViewTest(ViewTest):
    def __init__(self, *args, **kwargs):
        """Build a fleet."""
        super().__init__(*args, **kwargs)
        # create simple fleet
        config = EmptyBotConfig.objects.create(space=self.testcase_instance.space, settings={"empty": True})
        controller = Controller.get(
            exchange_account=self.testcase_instance.exchange_account,
            base="BTC",
            quote="USDT",
            interval="1m",
        )
        architecture = SinglePairArchitecture.objects.create(constants={"controller": controller})
        strategy = EmptyStrategy.objects.create(config=config, architecture=architecture)
        bot = Bot.objects.create(name="Test Bot", strategy=strategy)
        self.fleet = Fleet.objects.create(
            name="Test Fleet",
            exchange_account=self.testcase_instance.exchange_account,
            clusters=[
                {
                    "template_bot": bot,
                    "share": 0.7,
                    "breakpoint": 1000,
                    "autoscale": False,
                },
                {
                    "template_bot": bot,
                    "share": 0.3,
                    "breakpoint": 1000,
                    "autoscale": True,
                },
            ],
        )

    def check_response(self, response) -> bool:
        return compare_responses(
            response.data,
            {
                "uuid": str(),
                "name": str(),
                "created_at": str(),
                "statistics": {
                    "value": float(),
                    "order_count_30": float(),
                    "change_30": None,
                },
                "wallet": {},
                # "history": {},
                "bots": [],
            },
            authorize_abstract_comparison=True,
        )

    def run(self):
        print("retrieve run")
        return [
            {
                "name": "response",
                "setup": self.setup(),
                "status_code": 200,
                "checks": [self.check_response],
            },
        ]


class FleetViewTestCase:
    # def test_list(self):
    #     self.run_tests("list")

    # def test_retrieve(self):
    #     self.run_tests("list")

    def setup(self):
        list_test = ListFleetViewTest(self)
        # retrieve_test = RetrieveFleetViewTest(self)

        return {
            "list": {
                "url_name": "fleet-list",
                "view": FleetView,
                "method": "GET",
                "tests": list_test.run(),
            },
            # "retrieve": {
            #     "url_name": "fleet-detail",
            #     "view": FleetView,
            #     "method": "GET",
            #     "kwargs": {"pk": self.retrieve_test.fleet.uuid},
            #     "tests": retrieve_test.run(),
            # },
        }


class FleetViewAPIBinanceTestCase(FleetViewTestCase, APITestCase):
    exchange = "BINANCE"



from typing import TYPE_CHECKING

from django.db import models
from pydantic import BaseModel

if TYPE_CHECKING:
    from django_napse.core.models.wallets.wallet import Wallet


class CurrencyPydantic(BaseModel):
    """A Pydantic model for the Currency class."""

    ticker: str
    amount: float
    mbp: float


class Currency(models.Model):
    """A Currency contains the amount of a ticker in a wallet, as well as the Mean Buy Price (MBP)."""

    wallet = models.ForeignKey("Wallet", on_delete=models.CASCADE, related_name="currencies")
    mbp = models.FloatField()
    ticker = models.CharField(max_length=10)
    amount = models.FloatField(default=0)

    class Meta:  # noqa
        unique_together = ("wallet", "ticker")

    def __str__(self) -> str:  # pragma: no cover
        return f"CURRENCY {self.pk}"

    def info(self, beacon: str = "", *, verbose: bool = True) -> str:
        """Return a string with the history information.

        Args:
            beacon (str, optional): The prefix for each line. Defaults to "".
            verbose (bool, optional): Whether to print the string. Defaults to True.

        Returns:
            str: The string with the history information.
        """
        string = ""
        string += f"{beacon}Currency ({self.pk=}):\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.wallet=}\n"
        string += f"{beacon}\t{self.mbp=}\n"
        string += f"{beacon}\t{self.ticker=}\n"
        string += f"{beacon}\t{self.amount=}\n"

        if verbose:  # pragma: no cover
            print(string)
        return string

    @property
    def testing(self) -> bool:
        """Return the testing status of the wallet."""
        return self.wallet.testing

    def copy(self, owner: "Wallet") -> "Currency":
        """Return a copy of the currency with a new owner."""
        return Currency.objects.create(
            wallet=owner,
            mbp=self.mbp,
            ticker=self.ticker,
            amount=self.amount,
        )



# Generated by Django 4.2.5 on 2023-10-14 14:17

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("django_napse_auth", "0002_napseapikey"),
    ]

    operations = [
        migrations.AlterField(
            model_name="keypermission",
            name="key",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="permissions",
                to="django_napse_auth.napseapikey",
            ),
        ),
    ]



from typing import ClassVar

from django_napse.core.models import Controller, EmptyStrategy
from django_napse.utils.model_test_case import ModelTestCase
from tests.django_tests.db.bots.test_strategy import StrategyDefaultTestCase

"""
python tests/test_app/manage.py test tests.django_tests.bots.test_strategies.test_empty_bot -v2 --keepdb --parallel
"""


class EmptyBotTestCase(StrategyDefaultTestCase):
    model = EmptyStrategy
    config_settings: ClassVar = {"empty": True}

    @property
    def architecture_constants(self):
        return {
            "controller": Controller.get(
                exchange_account=self.exchange_account,
                base="BTC",
                quote="USDT",
                interval="1m",
            ),
        }


class EmptyBotBINANCETestCase(EmptyBotTestCase, ModelTestCase):
    exchange = "BINANCE"



from django.db import models

from django_napse.core.models.histories.history import History


class BotHistory(History):
    """A History for a Bot.

    Use it to track the evolution of a bot over time.

    This tracks the following fields:
    TODO
    """

    owner = models.OneToOneField("Bot", on_delete=models.CASCADE, related_name="history")

    def generate_data_points(self) -> None:
        """Create a new data point for the bot."""
        return



# Generated by Django 4.2.6 on 2023-10-23 19:44

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("django_napse_core", "0007_historydatapoint_created_at"),
        ("django_napse_auth", "0004_napseapikey_description"),
    ]

    operations = [
        migrations.AlterField(
            model_name="keypermission",
            name="space",
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name="permissions", to="django_napse_core.napsespace"),
        ),
    ]



from django.db import models


class Link(models.Model):
    bot = models.OneToOneField("Bot", on_delete=models.CASCADE, related_name="link")
    cluster = models.ForeignKey("Cluster", on_delete=models.CASCADE, related_name="links")
    importance = models.FloatField()

    def __str__(self):
        return f"LINK: {self.bot=} {self.cluster=}"

    def info(self, verbose=True, beacon=""):
        string = ""
        string += f"{beacon}Link {self.pk}:\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.bot=}\n"
        string += f"{beacon}\t{self.cluster=}\n"
        string += f"{beacon}\t{self.importance=}\n"

        if verbose:  # pragma: no cover
            print(string)
        return string



from rest_framework import serializers

from django_napse.api.bots.serializers.architecture_serializer import ArchitectureSerializer
from django_napse.api.bots.serializers.config_serializer import ConfigSerializer
from django_napse.core.models.bots.strategy import Strategy


class StrategySerializer(serializers.ModelSerializer):
    """Serialize a Strategy instance."""

    config = ConfigSerializer()
    architecture = ArchitectureSerializer()

    class Meta:  # noqa: D106
        model = Strategy
        fields = "__all__"



from datetime import datetime

from pytz import UTC

from django_napse.core.models import Controller
from django_napse.simulations.models import Candle, DataSet
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.datasets.test_candle -v2 --keepdb --parallel
"""


class CandleTestCase:
    model = Candle

    def simple_create(self):
        controller = Controller.get(
            exchange_account=self.exchange_account,
            base="BTC",
            quote="USDT",
            interval="1d",
        )
        dataset = DataSet.objects.create(
            controller=controller,
            start_date=datetime(2021, 1, 1, tzinfo=UTC),
            end_date=datetime(2021, 4, 1, tzinfo=UTC),
        )
        return Candle.objects.create(
            dataset=dataset,
            open_time=datetime(2020, 1, 1, tzinfo=UTC),
            open=1,
            high=2,
            low=0.5,
            close=1.5,
            volume=100,
        )


class CandleBINANCETestCase(CandleTestCase, ModelTestCase):
    exchange = "BINANCE"



from enum import EnumMeta, StrEnum
from typing import Iterator


class CustomEnumMeta(EnumMeta):
    """Custom EnumMeta class to allow string comparison for Enums."""

    def __contains__(self, obj: object) -> bool:
        """Check if obj is a str in Enum's value or if it's an Enum in Enum's members."""
        if isinstance(obj, str):
            return any(obj == item for item in self)
        return super().__contains__(obj)

    def __iter__(self) -> Iterator:
        """Allow to iterate over the Enum's values."""
        return (self._member_map_[name].value for name in self._member_names_)

    def __str__(self) -> str:
        """Return the Enum's value."""
        return f"{[self._member_map_[name].value for name in self._member_names_]}"


class EXCHANGES(StrEnum, metaclass=CustomEnumMeta):
    """The exchange for a fleet or a bot."""

    BINANCE = "BINANCE"


class TRANSACTION_TYPES(StrEnum, metaclass=CustomEnumMeta):  # noqa: N801, D101
    TRANSFER = "TRANSFER"
    CONNECTION_DEPOSIT = "CONNECTION_DEPOSIT"
    CONNECTION_WITHDRAW = "CONNECTION_WITHDRAW"
    ORDER_DEPOSIT = "ORDER_DEPOSIT"
    ORDER_PAYOUT = "ORDER_PAYOUT"
    ORDER_REFUND = "ORDER_REFUND"
    FLEET_REBALANCE = "FLEET_REBALANCE"


class ORDER_STATUS(StrEnum, metaclass=CustomEnumMeta):  # noqa: N801, D101
    PENDING = "PENDING"
    READY = "READY"
    PASSED = "PASSED"
    ONLY_BUY_PASSED = "ONLY_BUY_PASSED"
    ONLY_SELL_PASSED = "ONLY_SELL_PASSED"
    FAILED = "FAILED"


class SIDES(StrEnum, metaclass=CustomEnumMeta):  # noqa: D101
    BUY = "BUY"
    SELL = "SELL"
    KEEP = "KEEP"


class DOWNLOAD_STATUS(StrEnum, metaclass=CustomEnumMeta):  # noqa: N801, D101
    IDLE = "IDLE"
    DOWNLOADING = "DOWNLOADING"


class SIMULATION_STATUS(StrEnum, metaclass=CustomEnumMeta):  # noqa: N801, D101
    IDLE = "IDLE"
    RUNNING = "RUNNING"


class MODIFICATION_STATUS(StrEnum, metaclass=CustomEnumMeta):  # noqa: N801, D101
    PENDING = "PENDING"
    APPLIED = "APPLIED"
    REJECTED = "REJECTED"


class PLUGIN_CATEGORIES(StrEnum, metaclass=CustomEnumMeta):  # noqa: N801
    """The category for a plugin."""

    PRE_ORDER = "PRE_ORDER"
    POST_ORDER = "POST_ORDER"


class PERMISSION_TYPES(StrEnum, metaclass=CustomEnumMeta):  # noqa: N801
    """The permission type for a key."""

    ADMIN = "ADMIN"
    FULL_ACCESS = "FULL_ACCESS"
    READ_ONLY = "READ_ONLY"


class HISTORY_DATAPOINT_FIELDS(StrEnum, metaclass=CustomEnumMeta):  # noqa: N801
    """The different fields for a history data point."""

    WALLET_VALUE = "WALLET_VALUE"
    AMOUNT = "AMOUNT"
    ASSET = "ASSET"
    PRICE = "PRICE"
    MBP = "MBP"
    LBO = "LBO"
    VALUE = "VALUE"


class HISTORY_DATAPOINT_FIELDS_WILDCARDS(StrEnum, metaclass=CustomEnumMeta):  # noqa: N801
    """The different fields for a history data point."""

    AMOUNT = "AMOUNT_"


for wildcard in HISTORY_DATAPOINT_FIELDS_WILDCARDS:
    duplicate = False
    for field in HISTORY_DATAPOINT_FIELDS:
        if field.startswith(wildcard):
            duplicate = True
            error_msg = f"Duplicate field {field} for wildcard {wildcard}"
            raise ValueError(error_msg)

ORDER_LEEWAY_PERCENTAGE = 10

DEFAULT_TAX = {
    "BINANCE": 0.1,
}

EXCHANGE_TICKERS = {
    "BINANCE": ["BTC", "ETH", "USDT", "BNB", "XRP", "ADA", "DOGE", "MATIC", "SOL", "DOT", "LTC", "TRX", "SHIB", "AVAX", "LINK", "ATOM", "UNI", "XLM"],
}
EXCHANGE_PAIRS = {
    "BINANCE": {ticker + "USDT": {"base": ticker, "quote": "USDT"} for ticker in EXCHANGE_TICKERS["BINANCE"] if ticker != "USDT"},
}
EXCHANGE_INTERVALS = {
    "BINANCE": ("1m", "3m", "5m", "15m", "30m", "1h", "2h", "4h", "6h", "8h", "12h", "1d", "3d", "1w", "1M"),
}
EXCHANGE_TESTING = {
    "BINANCE": [True],
}
EXCHANGE_SIMULATION = {
    "BINANCE": True,
}

STABLECOINS = {
    "BINANCE": ["USDT", "DAI", "BUSD"],
}

EXCHANGE_CONSTANTS = {
    "EXCHANGES": list(EXCHANGES),
    "EXCHANGE_TICKERS": EXCHANGE_TICKERS,
    "EXCHANGE_PAIRS": EXCHANGE_PAIRS,
    "EXCHANGE_INTERVALS": EXCHANGE_INTERVALS,
    "EXCHANGE_TESTING": EXCHANGE_TESTING,
    "EXCHANGE_SIMULATION": EXCHANGE_SIMULATION,
    "STABLECOINS": STABLECOINS,
}


MONTH_NUM_TO_STR = {
    1: "Jan",
    2: "Feb",
    3: "Mar",
    4: "Apr",
    5: "May",
    6: "Jun",
    7: "Jul",
    8: "Aug",
    9: "Sep",
    10: "Oct",
    11: "Nov",
    12: "Dec",
}



from typing import TYPE_CHECKING, Optional

from django.db import models

from django_napse.core.models.orders.managers import OrderManager
from django_napse.core.models.transactions.credit import Credit
from django_napse.core.models.transactions.debit import Debit
from django_napse.core.models.transactions.transaction import Transaction
from django_napse.utils.constants import EXCHANGE_PAIRS, MODIFICATION_STATUS, ORDER_STATUS, SIDES, TRANSACTION_TYPES
from django_napse.utils.errors import OrderError

if TYPE_CHECKING:
    from django_napse.core.models.accounts.exchange import ExchangeAccount
    from django_napse.core.models.bots.controller import Controller
    from django_napse.core.models.modifications.modification import Modification


class OrderBatch(models.Model):
    """Represent a batch of bots' orders."""

    status = models.CharField(default=ORDER_STATUS.PENDING, max_length=15)
    controller: "Controller" = models.ForeignKey("Controller", on_delete=models.CASCADE, related_name="order_batches")

    def __str__(self) -> str:
        return f"ORDER BATCH {self.pk}"

    def set_status_ready(self) -> None:
        """Change status of the batch from PENDING to READY."""
        if self.status == ORDER_STATUS.PENDING:
            self.status = ORDER_STATUS.READY
            self.save()
        else:
            error_msg = f"Order {self.pk} is not pending."
            raise OrderError.StatusError(error_msg)

    def _set_status_post_process(self, receipt: dict) -> None:
        if self.status != ORDER_STATUS.READY:
            error_msg = f"Order {self.pk} is not ready."
            raise OrderError.StatusError(error_msg)
        buy_failed = False
        sell_failed = False
        if "error" in receipt[SIDES.BUY]:
            buy_failed = True
        if "error" in receipt[SIDES.SELL]:
            sell_failed = True
        if buy_failed and sell_failed:
            self.status = ORDER_STATUS.FAILED
        elif buy_failed:
            self.status = ORDER_STATUS.ONLY_SELL_PASSED
        elif sell_failed:
            self.status = ORDER_STATUS.ONLY_BUY_PASSED
        else:
            self.status = ORDER_STATUS.PASSED


class Order(models.Model):
    """An order market created by bots."""

    batch = models.ForeignKey("OrderBatch", on_delete=models.CASCADE, related_name="orders")
    connection = models.ForeignKey("Connection", on_delete=models.CASCADE, related_name="orders")
    price = models.FloatField()
    pair = models.CharField(max_length=10)
    side = models.CharField(max_length=10)
    completed = models.BooleanField(default=False)

    asked_for_amount = models.FloatField()
    asked_for_ticker = models.CharField(max_length=10)

    debited_amount = models.FloatField(default=0)

    batch_share = models.FloatField(default=0)
    exit_amount_base = models.FloatField(default=0)
    exit_amount_quote = models.FloatField(default=0)
    fees = models.FloatField(default=0)
    fee_ticker = models.CharField(max_length=10, blank=True)

    created_at = models.DateTimeField(auto_now_add=True)

    objects = OrderManager()

    def __str__(self) -> str:
        return f"ORDER: {self.pk=}"

    def info(self, beacon: str = "", *, verbose: bool = True) -> str:
        """Return a string with the model information.

        Args:
            beacon (str, optional): The prefix for each line. Defaults to "".
            verbose (bool, optional): Whether to print the string. Defaults to True.

        Returns:
            str: The string with the history information.
        """
        string = ""
        string += f"{beacon}Order ({self.pk=}):\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.connection=}\n"
        string += f"{beacon}\t{self.pair=}\n"
        string += f"{beacon}\t{self.asked_for_amount=}\n"
        string += f"{beacon}\t{self.asked_for_ticker=}\n"
        string += f"{beacon}\t{self.debited_amount=}\n"
        string += f"{beacon}\t{self.batch_share=}\n"
        string += f"{beacon}\t{self.exit_amount_base=}\n"
        string += f"{beacon}\t{self.exit_amount_quote=}\n"
        string += f"{beacon}\t{self.fees=}\n"
        string += f"{beacon}\t{self.fee_ticker=}\n"
        string += f"{beacon}\t{self.side=}\n"
        string += f"{beacon}\t{self.price=}\n"
        string += f"{beacon}\t{self.completed=}\n"

        new_beacon = beacon + "\t"
        string += f"{beacon}Wallet:\n"
        string += f"{beacon}{self.wallet.info(verbose=False, beacon=new_beacon)}\n"
        string += f"{beacon}Transactions in:\n"
        if self.wallet.transactions_to.all().count() == 0:
            string += f"{beacon}\tNone.\n"
        else:
            for transaction in self.wallet.transactions_to.all():
                trans_str = transaction.info(verbose=False, beacon=new_beacon)
                string += f"{trans_str}\n"
        string += f"{beacon}Transactions out:\n"
        if self.wallet.transactions_from.all().count() == 0:
            string += f"{beacon}\tNone.\n"
        else:
            for transaction in self.wallet.transactions_from.all():
                trans_str = transaction.info(verbose=False, beacon=new_beacon)
                string += f"{trans_str}\n"
        if verbose:  # pragma: no cover
            print(string)
        return string

    @property
    def testing(self) -> bool:
        """Whether the order is in testing mode.

        Returns:
            bool: Whether the order is in testing mode.
        """
        return self.connection.testing

    @property
    def exchange_account(self) -> "ExchangeAccount":
        """Return the exchange account of the order.

        Returns:
            ExchangeAccount: The exchange account of the order.
        """
        return self.connection.wallet.exchange_account

    def calculate_exit_amounts(self, controller: "Controller", executed_amounts: dict, fees: dict) -> None:
        """Calculate the exit amounts of the order.

        Exit amounts are the amounts that will be received by the user after the order is completed.

        Args:
            controller (Controller): The controller associated with the order.
            executed_amounts (dict): The amounts executed by the order.
            fees (dict): The fees paid by the user.

        Raises:
            OrderError.ProcessError: If the side of the order is not SELL or BUY.
        """
        if self.batch_share != 0:
            if self.side == SIDES.BUY:
                if executed_amounts == {}:
                    self.exit_amount_base = 0
                    self.exit_amount_quote = self.debited_amount
                    self.fees = 0
                    self.fee_ticker = controller.base
                else:
                    self.exit_amount_base = executed_amounts[controller.base] * self.batch_share
                    self.exit_amount_quote = self.debited_amount + executed_amounts[controller.quote] * self.batch_share
                    self.fees = fees[controller.base] * self.batch_share
                    self.fee_ticker = controller.base
            elif self.side == SIDES.SELL:
                if executed_amounts == {}:
                    self.exit_amount_base = self.debited_amount
                    self.exit_amount_quote = 0
                    self.fees = 0
                    self.fee_ticker = controller.quote
                else:
                    self.exit_amount_base = self.debited_amount + executed_amounts[controller.base] * self.batch_share
                    self.exit_amount_quote = executed_amounts[controller.quote] * self.batch_share
                    self.fees = fees[controller.quote] * self.batch_share
                    self.fee_ticker = controller.quote
            else:
                error_msg = f"Souldn't be calculating exit amount for order {self.pk} with side {self.side}."
                raise OrderError.ProcessError(error_msg)
        else:
            self.exit_amount_base = 0
            self.exit_amount_quote = 0
            self.fees = 0
            self.fee_ticker = controller.base

    def calculate_batch_share(self, total: float) -> None:
        """Set the batch share of the order.

        The batch share is the share of the total amount of the batch that the order represents.
        """
        self.batch_share = self.asked_for_amount / total

    def passed(self, batch: Optional[OrderBatch] = None) -> bool:
        """Whether the order passed or not.

        Args:
            batch (Optional[OrderBatch], optional): The batch to check. Defaults to None.

        Returns:
            bool: Whether the order passed or not.
        """
        batch = batch or self.batch
        return (self.side == SIDES.BUY and (batch.status in (ORDER_STATUS.PASSED, ORDER_STATUS.ONLY_BUY_PASSED))) or (
            self.side == SIDES.SELL and (batch.status in (ORDER_STATUS.PASSED, ORDER_STATUS.ONLY_SELL_PASSED))
        )

    def apply_modifications__no_db(
        self,
        batch: OrderBatch,
        modifications: list["Modification"],
        **kwargs: dict,
    ) -> tuple[list[models.Model], list["Modification"]]:
        """Apply the modifications to the order.

        Returns:
            all_modifications (list[Modification]): The modifications applied to the order.

        """
        all_modifications = []
        all_modified_objects = []
        if self.passed(batch):
            for modification in modifications:
                modified_object, modification_object = modification.apply__no_db(order=self, **kwargs)
                all_modifications.append(modification_object)
                all_modified_objects.append(modified_object)
        else:
            for modification in [modification for modification in modifications if modification.ignore_failed_order]:
                modified_object, modification_object = modification.apply__no_db(order=self, **kwargs)
                all_modifications.append(modification_object)
                all_modified_objects.append(modified_object)
            for modification in [modification for modification in modifications if not modification.ignore_failed_order]:
                modification.status = MODIFICATION_STATUS.REJECTED
                all_modifications.append(modification)
        return all_modifications, all_modified_objects

    def apply_modifications(self) -> list["Modification"]:
        """Apply the modifications to the order, and save them to the database.

        Returns:
            list[Modification]: The modifications applied to the order.
        """
        modifications, modified_objects = self.apply_modifications__no_db(
            batch=self.batch,
            modifications=[modification.find() for modification in self.modifications.all()],
            strategy=self.connection.bot.strategy.find(),
            architecture=self.connection.bot.architecture.find(),
            currencies=self.connection.wallet.to_dict().currencies,
        )
        for modification in modifications:
            modification.save()
        for modified_object in modified_objects:
            modified_object.save()
        return modifications

    def apply_swap(self) -> None:
        """Swap quote into base (BUY) or base into quote (SELL)."""
        if self.side == SIDES.BUY:
            Debit.objects.create(
                wallet=self.wallet,
                amount=self.debited_amount - self.exit_amount_quote,
                ticker=self.batch.controller.quote,
            )
            Credit.objects.create(
                wallet=self.wallet,
                amount=self.exit_amount_base,
                ticker=self.batch.controller.base,
            )
        elif self.side == SIDES.SELL:
            Debit.objects.create(
                wallet=self.wallet,
                amount=self.debited_amount - self.exit_amount_base,
                ticker=self.batch.controller.base,
            )
            Credit.objects.create(
                wallet=self.wallet,
                amount=self.exit_amount_quote,
                ticker=self.batch.controller.quote,
            )

    def process_payout(self) -> None:
        """Make a payout or a refund depending on the passed() status."""
        if self.side == SIDES.KEEP:
            return
        if self.passed():
            Transaction.objects.create(
                from_wallet=self.wallet,
                to_wallet=self.connection.wallet,
                amount=self.exit_amount_base,
                ticker=self.batch.controller.base,
                transaction_type=TRANSACTION_TYPES.ORDER_PAYOUT,
            )
            Transaction.objects.create(
                from_wallet=self.wallet,
                to_wallet=self.connection.wallet,
                amount=self.exit_amount_quote,
                ticker=self.batch.controller.quote,
                transaction_type=TRANSACTION_TYPES.ORDER_PAYOUT,
            )
        else:
            Transaction.objects.create(
                from_wallet=self.wallet,
                to_wallet=self.connection.wallet,
                amount=self.exit_amount_base,
                ticker=self.batch.controller.base,
                transaction_type=TRANSACTION_TYPES.ORDER_REFUND,
            )
            Transaction.objects.create(
                from_wallet=self.wallet,
                to_wallet=self.connection.wallet,
                amount=self.exit_amount_quote,
                ticker=self.batch.controller.quote,
                transaction_type=TRANSACTION_TYPES.ORDER_REFUND,
            )

    def tickers_info(self) -> dict[str, str]:
        """Give informations about received, spent & fee tickers."""
        spent_ticker = EXCHANGE_PAIRS[self.connection.space.exchange_account.exchange.name][self.pair]["base" if self.side == SIDES.SELL else "quote"]
        received_ticker = EXCHANGE_PAIRS[self.connection.space.exchange_account.exchange.name][self.pair][
            "quote" if self.side == SIDES.SELL else "base"
        ]
        fee_ticker = self.fee_ticker

        return {
            "spent_ticker": spent_ticker,
            "received_ticker": received_ticker,
            "fee_ticker": fee_ticker,
        }



from django_napse.utils.custom_test_case import CustomTestCase


class ModelTestCase(CustomTestCase):
    model = None
    exchange = None

    def simple_create(self):
        error_msg = "You must define a simple_create method."
        raise NotImplementedError(error_msg)

    def test_setup(self):
        skip, reason = self.skip_condition()
        if skip:
            self.skipTest(reason)
        if self.model is None:
            error_msg = f"ModelTestCase subclass {self.__class__.__name__} must define a model attribute."
            raise NotImplementedError(error_msg)

    def test_simple_creation(self):
        skip, reason = self.skip_condition()
        if skip:
            self.skipTest(reason)
        instance = self.simple_create()
        if instance is None:
            error_msg = "simple_create method must return the created instance."
            raise ValueError(error_msg)

    def test_info(self):
        skip, reason = self.skip_condition()
        if skip:
            self.skipTest(reason)
        instance = self.simple_create()
        instance.info(verbose=False)

    def test_str(self):
        skip, reason = self.skip_condition()
        if skip:
            self.skipTest(reason)
        instance = self.simple_create()
        instance.__str__()



import uuid
from datetime import datetime
from typing import Optional


def instance_check(target_type: any) -> callable:
    """Produce a callable to check if an instance is of a specific type."""

    def _instance_check(instance: object, target_type: any = target_type) -> bool:
        return isinstance(instance, target_type)

    return _instance_check


class Field:
    """Fields represent models' fields in the serializer."""

    validate = None
    # Define if the getter method takes the serializer as argument.
    getter_takes_serializer = False

    def __init__(
        self,
        *,
        default: Optional[any] = None,
        source: str | None = None,
        required: bool = False,
        **kwargs: dict[str, any],  # noqa: ARG002 (for DRF compatibility)
    ) -> None:
        """Define basic parameters of the field.

        Parameters:
            required: bool
                Define if the field is required. Only used for validation.
            source: str
                Define the source of the field in the model (ex: `exchange.name`).
        """
        self.required = required
        self.source = source
        self.default = default

    def to_value(self, value: any) -> any:
        """Overwrite this method for custom transformation on the serialized value."""
        return value

    def as_getter(self, serializer_field_name: str, serializer_cls: type) -> None:  # noqa: ARG002
        """Return a getter method for the field."""
        return


class StrField(Field):
    """Represent a string.

    Can be used for model.CharField, model.TextField, ...
    """

    to_value: callable = staticmethod(str)
    validate: callable = staticmethod(instance_check(str))


class IntField(Field):
    """Represent a integer."""

    to_value: callable = staticmethod(int)
    validate: callable = staticmethod(instance_check(int))


class FloatField(Field):
    """Represent a float."""

    to_value: callable = staticmethod(float)
    validate: callable = staticmethod(instance_check(float))


class BoolField(Field):
    """Represent a boolean."""

    to_value: callable = staticmethod(bool)
    validate: callable = staticmethod(instance_check(bool))


class UUIDField(Field):
    """Represent a uuid."""

    to_value: callable = staticmethod(uuid.UUID)
    validate: callable = staticmethod(instance_check(uuid.UUID))

    @staticmethod
    def to_value(value: uuid.UUID | str) -> str:
        """Format & return the value."""
        if not isinstance(value, uuid.UUID):
            return str(uuid.UUID(value))
        return str(value)


class DatetimeField(Field):
    """Represent a date."""

    validate: callable = instance_check(datetime)

    @staticmethod
    def to_value(value: datetime) -> str:
        """Format & return the value."""
        return value.strftime("%Y-%m-%d %H:%M:%S")


class MethodField(Field):
    """MethodField can be used to serialize complexe behaviours."""

    getter_takes_serializer = True

    # Avoir type serialization on data
    to_value = None
    validate = None

    def __init__(self, method_name: str | None = None, **kwargs: dict[str, any]) -> None:
        """Define the method field."""
        super().__init__(**kwargs)
        self.method_name = method_name

    def as_getter(self, serializer_field_name: str, serializer_cls: type) -> callable:
        """Get the (get_<field> | method_name) method from the serializer class."""
        return getattr(
            serializer_cls,
            self.method_name or f"get_{serializer_field_name}",
        )



from datetime import datetime

from pytz import UTC

from django_napse.core.models import Bot, Controller, EmptyBotConfig, EmptyStrategy, LBOPlugin, MBPPlugin, SBVPlugin, SinglePairArchitecture
from django_napse.simulations.models import SimulationQueue
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.db.simulations.test_simulation_queue -v2 --keepdb --parallel
"""


class SimulationQueueTestCase:
    model = SimulationQueue

    def simple_create(self):
        config = EmptyBotConfig.objects.create(space=self.space, settings={"empty": True})
        architecture = SinglePairArchitecture.objects.create(
            constants={
                "controller": Controller.get(
                    exchange_account=self.exchange_account,
                    base="BTC",
                    quote="USDT",
                    interval="1d",
                ),
            },
        )
        strategy = EmptyStrategy.objects.create(config=config, architecture=architecture)
        MBPPlugin.objects.create(strategy=strategy)
        LBOPlugin.objects.create(strategy=strategy)
        SBVPlugin.objects.create(strategy=strategy)
        bot = Bot.objects.create(name="Test Bot", strategy=strategy)
        return SimulationQueue.objects.create(
            space=self.space,
            bot=bot,
            start_date=datetime(2020, 1, 1, tzinfo=UTC),
            end_date=datetime(2020, 1, 5, tzinfo=UTC),
            investments={"USDT": 1000},
        )

    def test_quick_simulation(self):
        simulation_queue = self.simple_create()
        simulation_queue.run_quick_simulation(verbose=False)

    def test_irl_simulation(self):
        simulation_queue = self.simple_create()
        simulation_queue.run_irl_simulation(verbose=False)


class SimulationQueueBINANCETestCase(SimulationQueueTestCase, ModelTestCase):
    exchange = "BINANCE"



from typing import ClassVar

from rest_framework import serializers

from django_napse.core.models import Transaction


class TransactionSerializer(serializers.ModelSerializer):
    """Serialize a transaction instance."""

    operation_type = serializers.CharField(default="TRANSACTION")

    class Meta:  # noqa: D106
        model = Transaction
        fields: ClassVar[list[str]] = [
            "amount",
            "ticker",
            "operation_type",
            "created_at",
        ]
        read_only_fields = fields



from datetime import datetime

from pytz import UTC

from django_napse.core.models import Controller
from django_napse.simulations.models import DataSet
from django_napse.utils.errors import DataSetError
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.datasets.test_dataset -v2 --keepdb --parallel
"""


class DataSetTestCase:
    model = DataSet

    def simple_create(self):
        controller = Controller.get(
            exchange_account=self.exchange_account,
            base="BTC",
            quote="USDT",
            interval="1d",
        )
        return DataSet.objects.create(
            controller=controller,
            start_date=datetime(2021, 1, 1, tzinfo=UTC),
            end_date=datetime(2021, 4, 1, tzinfo=UTC),
        )

    def test_download_on_top(self):
        dataset = self.simple_create()
        DataSet.objects.create(
            controller=dataset.controller,
            start_date=datetime(2021, 1, 1, tzinfo=UTC),
            end_date=datetime(2021, 5, 1, tzinfo=UTC),
        )

    def test_info_small(self):
        controller = Controller.get(
            exchange_account=self.exchange_account,
            base="BTC",
            quote="USDT",
            interval="1d",
        )
        dataset = DataSet.objects.create(
            controller=controller,
            start_date=datetime(2021, 1, 1, tzinfo=UTC),
            end_date=datetime(2021, 1, 3, tzinfo=UTC),
        )
        dataset.info(verbose=False)

    def test_error_save_completion_lt_0(self):
        dataset = self.simple_create()
        dataset.completion = -1
        with self.assertRaises(DataSetError.InvalidSettings):
            dataset.save()

    def test_error_save_completion_gt_100(self):
        dataset = self.simple_create()
        dataset.completion = 101
        with self.assertRaises(DataSetError.InvalidSettings):
            dataset.save()

    def test_error_save_status_not_in_download_status(self):
        dataset = self.simple_create()
        dataset.status = "random status"
        with self.assertRaises(DataSetError.InvalidSettings):
            dataset.save()

    def test_error_already_downloading(self):
        dataset = self.simple_create()
        dataset.status = "DOWNLOADING"
        dataset.save()
        with self.assertRaises(DataSetError.InvalidSettings):
            dataset.set_downloading()

    def test_error_already_idle(self):
        dataset = self.simple_create()
        dataset.status = "IDLE"
        dataset.save()
        with self.assertRaises(DataSetError.InvalidSettings):
            dataset.set_idle()

    def test_is_finished(self):
        dataset = self.simple_create()
        self.assertTrue(dataset.is_finished())

        dataset.status = "DOWNLOADING"
        dataset.completion = 50
        dataset.save()
        self.assertFalse(dataset.is_finished())

        dataset.status = "DOWNLOADING"
        dataset.completion = 100
        dataset.save()
        self.assertFalse(dataset.is_finished())

        dataset.status = "IDLE"
        dataset.completion = 50
        dataset.save()
        self.assertFalse(dataset.is_finished())


class DataSetBINANCETestCase(DataSetTestCase, ModelTestCase):
    exchange = "BINANCE"



from datetime import datetime
from typing import ClassVar

from pytz import UTC

from django_napse.core.models import Bot, Controller, EmptyBotConfig, EmptyStrategy, SinglePairArchitecture
from django_napse.simulations.models import SimulationQueue

"""
python tests/test_app/manage.py test tests.django_tests.bots.test_plugin -v2 --keepdb --parallel
"""


class PluginDefaultTestCase:
    strategy_class = EmptyStrategy
    config_settings: ClassVar = {"empty": True}

    def simple_create(self):
        return self.model.objects.create(strategy=self.strategy)

    @property
    def config(self):
        return self.strategy_class.config_class().objects.create(space=self.space, settings=self.config_settings)

    @property
    def architecture(self):
        return self.strategy_class.architecture_class().objects.create(constants=self.architecture_constants)

    @property
    def strategy(self):
        return self.strategy_class.objects.create(config=self.config, architecture=self.architecture)

    @property
    def architecture_constants(self):
        return {
            "controller": Controller.get(
                exchange_account=self.exchange_account,
                base="BTC",
                quote="USDT",
                interval="1m",
            ),
        }

    def test_simulation(self):
        config = EmptyBotConfig.objects.create(space=self.space, settings={"empty": True})
        architecture = SinglePairArchitecture.objects.create(
            constants={
                "controller": Controller.get(
                    exchange_account=self.exchange_account,
                    base="BTC",
                    quote="USDT",
                    interval="1d",
                ),
            },
        )
        strategy = EmptyStrategy.objects.create(config=config, architecture=architecture)
        self.model.objects.create(strategy=strategy)
        bot = Bot.objects.create(name="Test Bot", strategy=strategy)
        simulation_queue = SimulationQueue.objects.create(
            space=self.space,
            bot=bot,
            start_date=datetime(2020, 1, 1, tzinfo=UTC),
            end_date=datetime(2020, 1, 5, tzinfo=UTC),
            investments={"USDT": 10000},
        )

        simulation_queue.run_quick_simulation(verbose=False)



# Generated by Django 4.2.6 on 2023-10-17 13:38

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_napse_core", "0004_spacehistory_fleethistory_exchangeaccounthistory"),
    ]

    operations = [
        migrations.AddField(
            model_name="exchangeaccount",
            name="default",
            field=models.BooleanField(default=False),
        ),
    ]



from django.db.transaction import atomic
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response

from django_napse.api.custom_permissions import HasAPIKey, HasMasterKey
from django_napse.api.custom_viewset import CustomViewSet
from django_napse.api.keys.serializers import NapseAPIKeySerializer
from django_napse.api.keys.serializers.key import NapseAPIKeySpaceSerializer
from django_napse.auth.models import NapseAPIKey
from django_napse.utils.constants import PERMISSION_TYPES


class KeyView(CustomViewSet):
    permission_classes = [HasMasterKey]
    serializer_class = NapseAPIKeySerializer

    def get_queryset(self):
        return NapseAPIKey.objects.all()

    def get_object(self):
        return NapseAPIKey.objects.get(prefix=self.kwargs["pk"])

    def get_permissions(self):
        match self.action:
            case "connect" | "possible_permissions":
                return [HasAPIKey()]
            case "retrieve":
                return [HasAPIKey()]
            case _:
                return super().get_permissions()

    def create(self, request):
        if "name" not in request.data:
            return Response({"error": "Missing name"}, status=status.HTTP_400_BAD_REQUEST)

        _, key = NapseAPIKey.objects.create_key(name=request.data["name"], description=request.data.get("description", ""))

        return Response({"key": key}, status=status.HTTP_201_CREATED)

    def retrieve(self, request, pk):
        key = self.get_object()
        if "space" not in request.query_params:
            serializer = NapseAPIKeySerializer(key)
            return Response(serializer.data, status=status.HTTP_200_OK)
        serializer = NapseAPIKeySpaceSerializer(
            instance=key,
            context={
                "space": request.query_params["space"],
            },
        )
        return Response(serializer.data, status=status.HTTP_200_OK)

    def list(self, request):
        keys = self.get_queryset()
        if "space" not in request.query_params:
            serializer = NapseAPIKeySerializer(keys, many=True)
            return Response(serializer.data, status=status.HTTP_200_OK)

        serializer = NapseAPIKeySpaceSerializer(
            keys,
            many=True,
            context={
                "space": request.query_params["space"],
            },
        )
        data = {"keys": []}
        for key in serializer.data:
            if key["permissions"]:
                data["keys"].append(key)
        master_key = NapseAPIKey.objects.get(is_master_key=True)
        serializer = NapseAPIKeySerializer(master_key)
        data["master_key"] = serializer.data
        return Response(data, status=status.HTTP_200_OK)

    def destroy(self, request, pk):
        key = self.get_object()
        key.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)

    def patch(self, request, pk):
        key = self.get_object()
        with atomic():
            if "name" in request.data:
                key.name = request.data["name"]
            if "description" in request.data:
                key.description = request.data["description"]
            if "permissions" in request.data:
                for permission in key.permissions.filter(space=self.get_space(request)):
                    permission.delete()
                for permission in request.data["permissions"]:
                    key.add_permission(self.get_space(request), permission)
            key.save()
            if request.data.get("revoked", False) and not key.is_master_key:
                key.revoke()
        serializer = NapseAPIKeySerializer(key)
        return Response(serializer.data, status=status.HTTP_200_OK)

    @action(detail=False, methods=["GET"])
    def possible_permissions(self, request):
        return Response(list(PERMISSION_TYPES), status=status.HTTP_200_OK)

    @action(detail=False, methods=["GET"])
    def connect(self, request):
        return Response(status=status.HTTP_204_NO_CONTENT)



# Generated by Django 4.2.5 on 2023-10-14 14:07

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("rest_framework_api_key", "0005_auto_20220110_1102"),
        ("django_napse_auth", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="NapseAPIKey",
            fields=[
                (
                    "apikey_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="rest_framework_api_key.apikey",
                    ),
                ),
                ("is_master_key", models.BooleanField(default=False)),
            ],
            options={
                "verbose_name": "API key",
                "verbose_name_plural": "API keys",
                "ordering": ("-created",),
                "abstract": False,
            },
            bases=("rest_framework_api_key.apikey",),
        ),
    ]



from django_napse.core.models import SpaceHistory
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.db.histories -v2 --keepdb --parallel
"""


class SpaceHistoryTestCase:
    model = SpaceHistory

    def simple_create(self):
        return self.model.objects.create(owner=self.space)

    def test_owner(self):
        self.assertEqual(self.simple_create().owner, self.space)

    def test_get_or_create(self):
        history1 = self.model.get_or_create(self.space)
        history2 = self.model.get_or_create(self.space)
        self.assertEqual(history1, history2)


class SpaceHistoryBINANCETestCase(SpaceHistoryTestCase, ModelTestCase):
    exchange = "BINANCE"



from django.db import models

from django_napse.core.models.histories.history import History


class ExchangeAccountHistory(History):
    owner = models.OneToOneField("ExchangeAccount", on_delete=models.CASCADE, related_name="history")



from __future__ import annotations

from typing import ClassVar

from rest_framework import serializers

from django_napse.api.histories.serializers.datapoint_field import HistoryDataPointFieldSerializer
from django_napse.core.models.histories.history import HistoryDataPoint


class HistoryDataPointSerializer(serializers.ModelSerializer):
    """Serialize a wallet instance."""

    fields = HistoryDataPointFieldSerializer(many=True)

    class Meta:  # noqa: D106
        model = HistoryDataPoint
        fields: ClassVar[list[str]] = ["created_at", "fields"]



import os
from unittest import mock

import celery
from celery.app import trace
from celery.signals import setup_logging
from django.conf import settings

# set the default Django settings module for the 'celery' program.
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "tests.test_app.settings")
celery_app = celery.Celery("django_napse")


# Using a string here means the worker doesn't have to serialize
# the configuration object to child processes.
# - namespace='CELERY' means all celery-related configuration keys
#   should have a `CELERY_` prefix.
celery_app.config_from_object(settings, namespace="CELERY")

# Gestion des logs ------------------------------------------------------------------
# Remove the trace from celery logs
old_info = trace.info
trace.info = mock.Mock()


# Configure celery logging
@setup_logging.connect
def config_loggers(*args: list, **kwargs: dict) -> None:  # noqa: ARG001
    """Take settings.LOGGING as config for celery loggers (especially for `formatters`)."""
    from logging.config import dictConfig

    from django.conf import settings

    dictConfig(settings.LOGGING)


# Remove the strategy from celery logs
def strategy_log_free(*args: list, **kwargs: dict) -> callable:
    """Remove the strategy from celery logs."""
    kwargs["info"] = mock.Mock()
    return celery.worker.strategy.default(*args[1:], **kwargs)


# Load task modules from all registered Django app configs.
celery_app.autodiscover_tasks()



from django.db import models

from django_napse.utils.errors import ControllerError


class ControllerManager(models.Manager):
    def create(self, **kwargs):
        bypass = kwargs.pop("bypass", False)
        if not bypass:
            error_msg = "Controller.object.create() is not recommended. Use Controller.get() instead."
            raise ControllerError.BypassError(error_msg)
        return super().create(**kwargs)



import contextlib

from django_napse.api.spaces import SpaceView
from django_napse.utils.api_test_case import APITestCase, ViewTest
from django_napse.utils.dict_comparison import compare_responses

"""
python tests/test_app/manage.py test tests.django_tests.api.spaces.test_space_view -v2 --keepdb --parallel
"""


class ListSpaceViewTest(ViewTest):
    def check_response(self, response) -> bool:
        return compare_responses(
            response.data[0],
            {
                "name": str(),
                "description": str(),
                "uuid": str(),
                "delta": float(),
                "value": float(),
                "exchange_account": str(),
            },
        )

    def run(self):
        return [
            {
                "name": "response",
                "setup": self.setup(data=None),
                "status_code": 200,
                "checks": [self.check_response],
            },
        ]


class RetrieveSpaceViewTest(ViewTest):
    def check_response(self, response) -> bool:
        return compare_responses(
            response.data,
            {
                "name": str(),
                "description": str(),
                "uuid": str(),
                "exchange_account": str(),
                "created_at": str(),
                "statistics": {
                    "value": float(),
                    "order_count_30": float(),
                    "delta_30": float(),
                },
                "wallet": {},
                "history": [],
                "fleets": [],
            },
            authorize_abstract_comparison=True,
        )

    def run(self):
        return [
            {
                "name": "missing_data",
                "setup": self.setup(),
                "status_code": 200,
                "checks": [self.check_response],
            },
        ]


class SpaceViewTestCase:
    # def test_list(self):
    #     self.run_tests("list")

    # def test_retrieve(self):
    #     self.run_tests("retrieve")

    def setUp(self):
        super().setUp()
        from django_napse.core.models import SpaceHistory

        with contextlib.suppress(Exception):
            SpaceHistory.objects.create(owner=self.space)

    def setup(self):
        list_test = ListSpaceViewTest(self)
        retrieve_test = RetrieveSpaceViewTest(self)

        return {
            "list": {
                "url_name": "space-list",
                "view": SpaceView,
                "method": "GET",
                "tests": list_test.run(),
            },
            "retrieve": {
                "url_name": "space-detail",
                "view": SpaceView,
                "method": "GET",
                "kwargs": {"pk": self.space.uuid},
                "tests": retrieve_test.run(),
            },
        }


class SpaceViewAPIBinanceTestCase(SpaceViewTestCase, APITestCase):
    exchange = "BINANCE"



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework_api_key.permissions import HasAPIKey

from django_napse.api.custom_permissions import HasSpace
from django_napse.api.custom_viewset import CustomViewSet
from django_napse.api.fleets.serializers import FleetDetailSerializer, FleetMoneyFlowSerializer, FleetSerializer
from django_napse.core.models import Fleet, Space

if TYPE_CHECKING:
    from uuid import UUID

    from django.db.models import QuerySet
    from rest_framework.request import Request

    from django_napse.utils.serializers import Serializer


class FleetView(CustomViewSet):
    """View of a fleet.

    Query parameters:
        space: uuid of the space to filter on.
        space_containers (bool): If True, list endpoint returns fleets for each space (default = True).
    """

    permission_classes: ClassVar[list] = [HasAPIKey, HasSpace]
    serializer_class = FleetSerializer

    def get_queryset(self) -> QuerySet[Fleet]:
        """Return all fleets or only the ones from a space."""
        space_uuid = self.request.query_params.get("space", None)
        if space_uuid is None:
            return Fleet.objects.all()
        self.get_space = Space.objects.get(uuid=space_uuid)
        return self.get_space.fleets

    def get_serializer_class(self) -> Serializer:
        """Return the serializer class for each action."""
        actions: dict = {
            "list": FleetSerializer,
            "retrieve": FleetDetailSerializer,
            "create": FleetSerializer,
            "invest": FleetMoneyFlowSerializer,
            "withdraw": FleetMoneyFlowSerializer,
        }
        result = actions.get(self.action)
        return result if result else super().get_serializer_class()

    def get_permissions(self) -> list:
        """Return permissions for each action."""
        match self.action:
            case "list" | "create":
                return [HasAPIKey()]
            case _:
                return super().get_permissions()

    def get_object(self) -> Fleet:
        """Return the fleet instance for detail endpoint."""
        uuid = self.kwargs.get("pk", None)
        if uuid is None:
            return super().get_object()
        return Fleet.objects.get(uuid=uuid)

    def _get_boolean_query_param(self, param: str) -> bool | None:
        """Return None if a boolean cannot be found."""
        if isinstance(param, bool):
            return param

        if not isinstance(param, str):
            return None

        match param.lower():
            case "true" | "1":
                return True
            case "false" | "0":
                return False
            case _:
                return None

    def list(self, request: Request) -> None:
        """Return all available fleets (depending of the keys permissions & space containerization)."""
        space_containers = self._get_boolean_query_param(request.query_params.get("space_containers", True))
        space_uuid = request.query_params.get("space", None)
        api_key = self.get_api_key(request)

        if not space_containers and api_key.is_master_key:
            # Not space_containers mode is only available for master key
            serializer = self.get_serializer(self.get_queryset(), many=True)
            return Response(serializer.data, status=status.HTTP_200_OK)

        # Get spaces from API key
        spaces = Space.objects.all() if api_key.is_master_key else [permission.space for permission in api_key.permissions.all()]
        # Filter by specific space
        if space_uuid is not None:
            space = Space.objects.get(uuid=space_uuid)
            if space not in spaces:
                return Response(status=status.HTTP_403_FORBIDDEN)
            spaces = [space]

        # Fleet list
        fleets = []
        for space in spaces:
            serializer = self.get_serializer(space.fleets, many=True, space=space)
            if serializer.data != []:
                fleets += serializer.data
        return Response(fleets, status=status.HTTP_200_OK)

    def retrieve(self, request: Request, pk: str | int | UUID | None = None) -> None:  # noqa: ARG002
        """Return the detail info of the given fleet."""
        instance = self.get_object()
        space = self.get_space(request)
        serializer = self.get_serializer(instance, space=space)
        return Response(serializer.data, status=status.HTTP_200_OK)

    def create(self, request: Request) -> None:
        """Create a new fleet."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        fleet = serializer.save()
        space = serializer.space
        fleet.invest(space, 0, "USDT")
        return Response(serializer.data, status=status.HTTP_201_CREATED)

    def delete(self, request: Request, pk: str | int | UUID | None = None) -> None:  # noqa: ARG002
        """Delete a fleet."""
        return Response(status=status.HTTP_501_NOT_IMPLEMENTED)

    @action(detail=True, methods=["GET", "POST"])
    def invest(self, request: Request, pk: str | int | UUID | None = None) -> None:  # noqa: ARG002
        """Invest in a fleet."""
        fleet = self.get_object()
        space = self.get_space(request)

        if not space.testing:
            error_msg: str = "Investing in real is not allowed yet."
            return Response(error_msg, status=status.HTTP_403_FORBIDDEN)

        match request.method:
            case "POST":
                serializer = self.get_serializer(
                    data=request.data,
                    instance=fleet,
                    space=space,
                    side="INVEST",
                )
                serializer.is_valid(raise_exception=True)
                serializer.save()
                return Response(status=status.HTTP_200_OK)

            case "GET":
                from pprint import pprint

                pprint(space.wallet.currencies.all())
                possible_investments = [{"ticker": currency.ticker, "amout": currency.amount} for currency in space.wallet.currencies.all()]
                return Response(possible_investments, status=status.HTTP_200_OK)

            case _:
                return Response(status=status.HTTP_405_METHOD_NOT_ALLOWED)

    @action(detail=True, methods=["GET", "POST"])
    def withdraw(self, request: Request, pk: str | int | UUID | None = None) -> None:  # noqa: ARG002
        """Withdraw from a fleet."""
        return Response(status=status.HTTP_501_NOT_IMPLEMENTED)



import time
from datetime import datetime
from typing import TYPE_CHECKING

from django.db import models
from pydantic import BaseModel

from django_napse.core.models.bots.controller import Controller
from django_napse.core.models.wallets.currency import Currency, CurrencyPydantic
from django_napse.core.models.wallets.managers import WalletManager
from django_napse.utils.errors import WalletError
from django_napse.utils.findable_class import FindableClass

if TYPE_CHECKING:
    from django_napse.core.models.accounts.exchange import ExchangeAccount
    from django_napse.core.models.accounts.space import Space


class WalletPydantic(BaseModel):
    """A Pydantic model for the Wallet class."""

    title: str
    testing: bool
    locked: bool
    created_at: datetime
    currencies: dict[str, CurrencyPydantic]


class Wallet(models.Model, FindableClass):
    """A Wallet is a collection of currencies."""

    title = models.CharField(max_length=255, default="Wallet")
    locked = models.BooleanField(default=False)
    created_at = models.DateTimeField(auto_now_add=True)

    objects = WalletManager()

    def __str__(self) -> str:  # pragma: no cover
        return f"WALLET: {self.pk=}"

    def info(self, beacon: str = "", *, verbose: bool = True) -> str:
        """Return a string with the model information.

        Args:
            beacon (str, optional): The prefix for each line. Defaults to "".
            verbose (bool, optional): Whether to print the string. Defaults to True.

        Returns:
            str: The string with the history information.
        """
        self = self.find()
        string = ""
        string += f"{beacon}Wallet ({self.pk=}):\t{type(self)}\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.title=}\n"
        string += f"{beacon}\t{self.testing=}\n"
        string += f"{beacon}\t{self.locked=}\n"
        string += f"{beacon}Currencies\n"
        if self.currencies.count() == 0:
            string += f"{beacon}\tNo currencies\n"
        else:
            for currency in self.currencies.all():
                string += f"{beacon}\t{currency.ticker}: {currency.amount} @ ${currency.mbp}\n"
        if verbose:  # pragma: no cover
            print(string)
        return string

    @property
    def testing(self) -> bool:
        """Return whether the wallet is in testing mode."""
        return self.owner.testing

    @property
    def space(self) -> "Space":  # pragma: no cover
        """Return the space that owns the wallet."""
        error_msg = f"space() not implemented by default. Please implement it in {self.__class__}."
        raise NotImplementedError(error_msg)

    @property
    def exchange_account(self) -> "ExchangeAccount":  # pragma: no cover
        """Return the exchange account that contains the wallet."""
        error_msg = "exchange_account() not implemented by default. Please implement in a subclass of Wallet."
        raise NotImplementedError(error_msg)

    def spend(self, amount: float, ticker: str, recv: int = 3, **kwargs: dict) -> None:
        """Spend an amount of a currency from the wallet.

        Args:
            amount (float): The amount to spend.
            ticker (str): The ticker of the currency to spend.
            recv (int, optional): The time to wait before failing the transaction. Defaults to 3.
            kwargs (dict): Additional arguments.

        Raises:
            WalletError.SpendError: If you try to spend money from the wallet without using the Transactions class.
            ValueError: If the amount is negative.
            TimeoutError: If the wallet is locked for too long.
            WalletError.SpendError: If the currency does not exist in the wallet.
            WalletError.SpendError: If there is not enough money in the wallet.
        """
        if not kwargs.get("force", False):
            error_msg = "DANGEROUS: You should not use this method outside of select circumstances. Use Transactions instead."
            raise WalletError.SpendError(error_msg)

        if amount <= 0:
            error_msg: str = f"Amount must be positive, got {amount}"
            raise ValueError(error_msg)

        start_time = time.time()
        while self.locked:
            time.sleep(0.01)
            if time.time() - start_time > recv:
                error_msg: str = f"Wallet: {self.title} is locked"
                raise TimeoutError(error_msg)
        self.locked = True
        self.save()

        try:
            currency = self.currencies.get(ticker=ticker)
        except Currency.DoesNotExist as err:
            self.locked = False
            self.save()
            error_msg: str = f"Currency {ticker} does not exist in wallet: {self.title}."
            raise WalletError.SpendError(error_msg) from err

        if currency.amount < amount:
            self.locked = False
            self.save()
            error_msg: str = f"Not enough money in wallet: {self.title} ({amount} > {currency.amount} {ticker})."
            raise WalletError.SpendError(error_msg)

        currency.amount -= amount
        currency.save()
        self.locked = False
        self.save()

    def top_up(self, amount: float, ticker: str, mbp: float | None = None, recv: int = 3, **kwargs: dict) -> None:
        """Top up the wallet with an amount of a currency.

        Args:
            amount (float): The amount to top up.
            ticker (str): The ticker of the currency to top up.
            mbp (float, optional): The price of the currency. Defaults to None.
            recv (int, optional): The time to wait before failing the transaction. Defaults to 3.
            kwargs (dict): Additional arguments.

        Raises:
            WalletError.TopUpError: If you try to top up the wallet without using the Transactions class.
            ValueError: If the amount is negative.
            TimeoutError: If the wallet is locked for too long.
        """
        if not kwargs.get("force", False):
            error_msg = "DANGEROUS: You should not use this method outside of select circumstances. Use Transactions instead."
            raise WalletError.TopUpError(error_msg)

        if amount <= 0:
            error_msg: str = f"Amount must be positive, got {amount}"
            raise ValueError(error_msg)

        if mbp is None:
            mbp = Controller.get_asset_price(exchange_account=self.exchange_account, base=ticker)

        start_time = time.time()
        while self.locked:
            time.sleep(0.01)
            if time.time() - start_time > recv:
                error_msg: str = f"Wallet: {self.title} is locked"
                raise TimeoutError(error_msg)

        self.locked = True
        self.save()
        try:
            currency = self.currencies.get(ticker=ticker)
        except Currency.DoesNotExist:
            currency = Currency.objects.create(wallet=self, ticker=ticker, amount=0, mbp=0)
        currency.mbp = (currency.mbp * currency.amount + mbp * amount) / (currency.amount + amount)
        currency.amount += amount
        currency.save()
        self.locked = False
        self.save()

    def has_funds(self, amount: float, ticker: str) -> bool:
        """Check if the wallet has enough funds.

        Args:
            amount (float): The amount the wallet should have.
            ticker (str): The ticker of the currency.

        Returns:
            bool: Whether the wallet has enough funds.
        """
        try:
            currency = self.currencies.get(ticker=ticker)
        except Currency.DoesNotExist:
            return False
        return currency.amount >= amount

    def get_amount(self, ticker: str) -> float:
        """Return the amount of a currency in the wallet.

        Args:
            ticker (str): The ticker of the currency.

        Returns:
            float: The amount of the currency in the wallet.
        """
        try:
            curr = self.currencies.get(ticker=ticker)
        except Currency.DoesNotExist:
            return 0
        return curr.amount

    def value_mbp(self) -> float:
        """Return the value of the wallet in USDT.

        Returns:
            float: The value of the wallet in USDT.
        """
        value = 0
        for currency in self.currencies.all():
            if currency.amount == 0:
                continue
            value += currency.amount * currency.mbp
        return value

    def value_market(self) -> float:
        """Return the value of the wallet in USDT.

        Returns:
            float: The value of the wallet in USDT.
        """
        value = 0
        for currency in self.currencies.all():
            if currency.amount == 0:
                continue
            value += currency.amount * Controller.get_asset_price(exchange_account=self.exchange_account, base=currency.ticker)
        return value

    def to_dict(self) -> WalletPydantic:
        """Return a dictionary representation of the wallet.

        Returns:
            dict: The dictionary representation of the wallet.
        """
        currencies = self.currencies.all()

        return WalletPydantic(
            title=self.title,
            testing=self.testing,
            locked=self.locked,
            created_at=self.created_at,
            currencies={
                currency.ticker: CurrencyPydantic(
                    ticker=currency.ticker,
                    amount=currency.amount,
                    mbp=currency.mbp,
                )
                for currency in currencies
            },
        )



from django.db import models

from django_napse.core.models.bots.config import BotConfig


class TurboDCABotConfig(BotConfig):
    setting_timeframe = models.DurationField()
    setting_percentage = models.FloatField()

    def __str__(self) -> str:
        return f"TURBO DCA BOT CONFIG: {self.pk} - {self.immutable}"



from .base_task import BaseTask
from .candle_collector import CandleCollectorTask
from .controller_update import ControllerUpdateTask
from .history_update import HistoryUpdateTask
from .order_process_executor import OrderProcessExecutorTask



import uuid

from django.db import models

from django_napse.utils.constants import PERMISSION_TYPES


class KeyPermission(models.Model):
    uuid = models.UUIDField(unique=True, editable=False, default=uuid.uuid4)

    key = models.ForeignKey("NapseAPIKey", on_delete=models.CASCADE, related_name="permissions")
    space = models.ForeignKey("django_napse_core.Space", on_delete=models.CASCADE, related_name="permissions")
    approved = models.BooleanField(default=False)
    revoked = models.BooleanField(default=False)
    permission_type = models.CharField(max_length=200)

    class Meta:
        unique_together = ["key", "space", "permission_type"]

    def __str__(self):
        return f"NAPSE KEY PERMISSION: {self.permission_type}"

    def save(self, *args, **kwargs):
        if self.permission_type not in PERMISSION_TYPES:
            error_msg = f"Permission type ({self.permission_type}) is not valid. Valid permission types are: {PERMISSION_TYPES}"
            raise ValueError(error_msg)
        super().save(*args, **kwargs)

    def info(self, verbose=True, beacon=""):
        string = ""
        string += f"{beacon}Napse Key Permission ({self.pk=}):\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.key=}\n"
        string += f"{beacon}\t{self.space=}\n"
        string += f"{beacon}\t{self.permission_type=}\n"
        string += f"{beacon}\t{self.valid=}\n"
        if verbose:  # pragma: no cover
            print(string)
        return string



from rest_framework.exceptions import APIException


class APIError:
    """Base class for API errors."""

    class MissingSpace(APIException):
        status_code = 400
        default_detail = "Missing space uuid in request data"
        default_code = "missing_space"

    class InvalidSpace(APIException):
        status_code = 400
        default_detail = "Space uuid is invalid"
        default_code = "invalid_space"

    class InvalidPermissions(APIException):
        status_code = 403
        default_detail = "API key does not have the required permissions."
        default_code = "invalid_permissions"

    class NoAPIKey(APIException):
        status_code = 403
        default_detail = "No API key was provided. Please provide an API key in the Authorization header. (Authorization: Api-Key <key>)"
        default_code = "no_api_key"

    class InvalidAPIKey(APIException):
        status_code = 403
        default_detail = "Invalid API key."
        default_code = "invalid_api_key"



"""Module that contains a function to compare two dictionaries."""
from rest_framework.utils.serializer_helpers import ReturnDict, ReturnList


def compare_responses(response1, response2, **kwargs):
    """Compare two responses.

    Kwargs:
    ------
    authorize_abstract_comparison (bool): Comparizon between an empty dict and a no empty dict is True. Same for list (default: False).
    """
    response1 = convert_response(response1)
    response2 = convert_response(response2)

    if response1 is None and response2 is None:
        return True
    if isinstance(response1, str) and isinstance(response2, str):
        return True
    if isinstance(response1, (int, float)) and isinstance(response2, (int, float)):
        return True
    if isinstance(response1, bool) and isinstance(response2, bool):
        return True
    if isinstance(response1, dict) and isinstance(response2, dict):
        return compare_dicts(response1, response2, **kwargs)
    if isinstance(response1, list) and isinstance(response2, list):
        return compare_list(response1, response2, **kwargs)
    if type(response1) != type(response2):
        error_msg: str = f"{type(response1)} != {type(response2)}"
        raise TypeError(error_msg)
    return True


def convert_response(response: ReturnList):
    if isinstance(response, ReturnList):
        return list(response)
    if isinstance(response, ReturnDict):
        return dict(response)
    return response


def compare_dicts(dict1, dict2, **kwargs):
    if not isinstance(dict1, dict) or not isinstance(dict2, dict):
        error_message = "One of the arguments is not a dictionary."
        raise TypeError(error_message)

    # Only dict type comparaison
    if kwargs.get("authorize_abstract_comparison", False) and (dict1 == {} or dict2 == {}):
        return True

    if set(dict1.keys()) != set(dict2.keys()):
        error_msg: str = f"Dictionaries do not contain the same keys ({len(dict1.keys())} != {len(dict2.keys())})\n"
        error_msg += f"{dict1.keys()} != {dict2.keys()}"
        raise ValueError(error_msg)

    for key in dict1:
        val1 = dict1[key]
        val2 = dict2[key]
        try:
            compare_responses(val1, val2, **kwargs)
        except TypeError as error:
            error_msg = f"{error} on {key} key"
            raise TypeError(error_msg) from error
    return True


def compare_list(list1, list2, **kwargs):
    if not isinstance(list1, list) or not isinstance(list2, list):
        error_message = "One of the arguments is not a list."
        raise TypeError(error_message)

    # Only list type comparaison
    if kwargs.get("authorize_abstract_comparison", False) and (list1 == [] or list2 == []):
        return True

    if len(list1) != len(list2):
        error_msg: str = f"List haven't the same lenght ({len(list1)} != {len(list1)})"
        raise ValueError(error_msg)

    for i in range(len(list1)):
        try:
            compare_responses(list1[i], list2[i], **kwargs)
        except TypeError as error:
            error_msg = f"{error} on {i} index"
            raise TypeError(error_msg) from error
    return True



from django.apps import apps
from django.db import models

from django_napse.utils.errors import SimulationError


class SimulationManager(models.Manager):
    def create(self, space, bot, start_date, end_date, simulation_reference, data):
        SimulationDataPoint = apps.get_model("django_napse_simulations", "SimulationDataPoint")
        SimulationDataPointExtraInfo = apps.get_model("django_napse_simulations", "SimulationDataPointExtraInfo")

        simulation = self.model(
            space=space,
            bot=bot,
            start_date=start_date,
            end_date=end_date,
            simulation_reference=simulation_reference,
        )
        simulation.save()

        if data == {}:
            return simulation
        must_have = ["dates", "values", "actions", "amounts", "tickers"]
        for key in must_have:
            if key not in data:
                error_msg = f"Key {key} not in data"
                raise ValueError(error_msg)

        first_length = len(data["dates"])
        for key, value in data.items():
            if len(value) != first_length:
                error_msg = f"Key {key} has length {len(value)} but should have length {first_length}"
                raise SimulationError.InvalidData(error_msg)

        extra_info_keys = [key for key in data if key not in must_have]

        simulation.save()
        bulk_list_data_point = []
        bulk_list_extra_info = []
        for info in zip(
            data["dates"],
            data["values"],
            data["actions"],
            data["amounts"],
            data["tickers"],
            *[data[key] for key in extra_info_keys],
            strict=True,
        ):
            date = info[0]
            value = info[1]
            action = info[2]
            amount = info[3]
            ticker = info[4]
            extra_info = {key: info[i + 5] for i, key in enumerate(extra_info_keys)}
            bulk_list_data_point.append(
                SimulationDataPoint(
                    simulation=simulation,
                    date=date,
                    value=value,
                    action=action,
                    amount=amount,
                    ticker=ticker,
                ),
            )
            for key, value in extra_info.items():
                bulk_list_extra_info.append(
                    SimulationDataPointExtraInfo(
                        data_point=bulk_list_data_point[-1],
                        key=key,
                        value=str(value),
                        target_type=type(value).__name__,
                    ),
                )

            if len(bulk_list_data_point) == 1000 or date == data["dates"][-1]:
                SimulationDataPoint.objects.bulk_create(bulk_list_data_point)
                bulk_list_data_point = []
                SimulationDataPointExtraInfo.objects.bulk_create(bulk_list_extra_info)
                bulk_list_extra_info = []

        return simulation



import math
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Optional

from django.db import models
from requests.exceptions import ConnectionError, ReadTimeout, SSLError

from django_napse.core.models.bots.managers.controller import ControllerManager
from django_napse.core.models.orders.order import Order, OrderBatch
from django_napse.utils.constants import EXCHANGE_INTERVALS, EXCHANGE_PAIRS, ORDER_STATUS, SIDES, STABLECOINS
from django_napse.utils.errors import ControllerError
from django_napse.utils.trading.binance_controller import BinanceController, ExchangeController

if TYPE_CHECKING:
    from django_napse.core.models.accounts.exchange import Exchange, ExchangeAccount
    from django_napse.core.models.bots.bot import Bot


class Controller(models.Model):
    """Model to control the trading of a pair on an exchange."""

    exchange_account: "ExchangeAccount" = models.ForeignKey("ExchangeAccount", on_delete=models.CASCADE, related_name="controller")

    pair = models.CharField(max_length=10)
    base = models.CharField(max_length=10)
    quote = models.CharField(max_length=10)
    interval = models.CharField(max_length=10)

    min_notional = models.FloatField(null=True)
    min_trade = models.FloatField(null=True)
    lot_size = models.IntegerField(null=True)
    price = models.FloatField(null=True)
    last_price_update = models.DateTimeField(null=True)
    last_settings_update = models.DateTimeField(null=True)

    objects = ControllerManager()

    class Meta:  # noqa: D106
        unique_together = ("pair", "interval", "exchange_account")

    def __str__(self) -> str:
        return f"Controller {self.pair} (id: {self.pk})"

    def save(self, *args: list[any], **kwargs: dict[str, any]) -> None:
        """Save the instance."""
        if not self.pk:
            self.base = self.base.upper()
            self.quote = self.quote.upper()
            if self.base == self.quote:
                error_msg = f"Base and quote cannot be the same: {self.base} - {self.quote}"
                raise ControllerError.InvalidSetting(error_msg)
            self.pair = self.base + self.quote
            self.update_variables_always()
            if self.interval not in EXCHANGE_INTERVALS[self.exchange.name]:
                error_msg = f"Invalid interval: {self.interval}"
                raise ControllerError.InvalidSetting(error_msg)
        return super().save(*args, **kwargs)

    def info(self, beacon: str = "", *, verbose: bool = True) -> str:
        """Return a string with the model information.

        Args:
            beacon (str, optional): The prefix for each line. Defaults to "".
            verbose (bool, optional): Whether to print the string. Defaults to True.

        Returns:
            str: The string with the history information.
        """
        string = ""
        string += f"{beacon}Controller {self.pk}:\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.pair=}\n"
        string += f"{beacon}\t{self.base=}\n"
        string += f"{beacon}\t{self.quote=}\n"
        string += f"{beacon}\t{self.interval=}\n"
        string += f"{beacon}\t{self.min_notional=}\n"
        string += f"{beacon}\t{self.min_trade=}\n"
        string += f"{beacon}\t{self.lot_size=}\n"
        string += f"{beacon}\t{self.price=}\n"
        string += f"{beacon}\t{self.last_price_update=}\n"
        string += f"{beacon}\t{self.last_settings_update=}\n"

        if verbose:  # pragma: no cover
            print(string)
        return string

    @staticmethod
    def get(exchange_account: "ExchangeAccount", base: str, quote: str, interval: str = "1m") -> "Controller":
        """Return a controller object from the database."""
        try:
            controller = Controller.objects.get(
                exchange_account=exchange_account,
                base=base,
                quote=quote,
                interval=interval,
            )
        except Controller.DoesNotExist:
            controller = Controller.objects.create(
                exchange_account=exchange_account,
                base=base,
                quote=quote,
                interval=interval,
                bypass=True,
            )
        controller.update_variables()
        return controller

    @property
    def exchange_controller(self) -> "ExchangeController":
        """Return the exchange controller of the controller.

        Returns:
            ExchangeController: The exchange controller of the controller.
        """
        return self.exchange_account.find().exchange_controller()

    @property
    def exchange(self) -> "Exchange":
        """Return the exchange of the controller."""
        return self.exchange_account.exchange

    def update_variables(self) -> None:
        """If the variables are older than 1 minute, update them."""
        if self.last_settings_update is None or self.last_settings_update < datetime.now(tz=timezone.utc) - timedelta(minutes=1):
            self.update_variables_always()

    def update_variables_always(self) -> None:
        """Update the variables of the controller."""
        exchange_controller = self.exchange_controller

        if exchange_controller == "DEFAULT":
            pass
        elif exchange_controller.__class__ == BinanceController:
            try:
                filters = exchange_controller.client.get_symbol_info(self.pair)["filters"]
                last_price = exchange_controller.client.get_ticker(symbol=self.pair)["lastPrice"]
                for specific_filter in filters:
                    if specific_filter["filterType"] == "LOT_SIZE":
                        self.lot_size = -int(math.log10(float(specific_filter["stepSize"])))
                    if specific_filter["filterType"] in ["MIN_NOTIONAL", "NOTIONAL"]:
                        self.min_notional = float(specific_filter["minNotional"])
                self.min_trade = self.min_notional / float(last_price)
                self.last_settings_update = datetime.now(tz=timezone.utc)
            except ReadTimeout:
                print("ReadTimeout in update_variables_always")
            except SSLError:
                print("SSLError in update_variables_always")
            except ConnectionError:
                print("ConnectionError in update_variables_always")
        else:
            error_msg = f"Exchange controller not supported: {exchange_controller.__class__}"
            raise NotImplementedError(error_msg)

        if self.pk:
            self.save()

    def process_orders(self, no_db_data: Optional[dict] = None, *, testing: bool) -> list[Order]:
        in_simulation = no_db_data is not None
        no_db_data = no_db_data or {
            "buy_orders": Order.objects.filter(
                order__batch__status=ORDER_STATUS.READY,
                order__side=SIDES.BUY,
                order__batch__controller=self,
                order__testing=testing,
            ),
            "sell_orders": Order.objects.filter(
                order__batch__status=ORDER_STATUS.READY,
                order__side=SIDES.SELL,
                order__batch__controller=self,
                order__testing=testing,
            ),
            "keep_orders": Order.objects.filter(
                order__batch__status=ORDER_STATUS.READY,
                order__side=SIDES.KEEP,
                order__batch__controller=self,
                order__testing=testing,
            ),
            "batches": OrderBatch.objects.filter(status=ORDER_STATUS.READY, batch__controller=self),
            "exchange_controller": self.exchange_controller,
            "min_trade": self.min_trade,
            "price": self.get_price(),
        }

        aggregated_order = {
            "buy_amount": 0,
            "sell_amount": 0,
            "min_trade": no_db_data["min_trade"],
            "price": no_db_data["price"],
            "min_notional": self.min_notional,
            "pair": self.pair,
        }

        for order in no_db_data["buy_orders"]:
            aggregated_order["buy_amount"] += order.asked_for_amount
        for order in no_db_data["sell_orders"]:
            aggregated_order["sell_amount"] += order.asked_for_amount

        if aggregated_order["buy_amount"] > 0:
            for order in no_db_data["buy_orders"]:
                order.calculate_batch_share(total=aggregated_order["buy_amount"])
        if aggregated_order["sell_amount"] > 0:
            for order in no_db_data["sell_orders"]:
                order.calculate_batch_share(total=aggregated_order["sell_amount"])
        for order in no_db_data["keep_orders"]:
            order.batch_share = 0

        receipt, executed_amounts_buy, executed_amounts_sell, fees_buy, fees_sell = no_db_data["exchange_controller"].submit_order(
            controller=self,
            aggregated_order=aggregated_order,
            testing=in_simulation or testing,
        )
        all_orders = []
        for order in no_db_data["buy_orders"]:
            order.calculate_exit_amounts(
                controller=self,
                executed_amounts=executed_amounts_buy,
                fees=fees_buy,
            )
            all_orders.append(order)
        for order in no_db_data["sell_orders"]:
            order.calculate_exit_amounts(
                controller=self,
                executed_amounts=executed_amounts_sell,
                fees=fees_sell,
            )
            all_orders.append(order)
        for order in no_db_data["keep_orders"]:
            order.calculate_exit_amounts(
                controller=self,
                executed_amounts={},
                fees={},
            )
            all_orders.append(order)

        for batch in no_db_data["batches"]:
            batch._set_status_post_process(receipt=receipt)

        return all_orders

    def apply_orders(self, orders):
        for order in orders:
            order.save()
            order.apply_swap()

    def send_candles_to_bots(self, closed_candle, current_candle) -> list:
        """Scan all bots (that are allowed to trade) and get their orders.

        Args:
        ----
        closed_candle : The candle that just closed.
        current_candle : The candle that is currently open.

        Returns:
        -------
        list: A list of orders.
        """
        orders = []
        for bot in self.bots.all().filter(is_simulation=False, fleet__running=True, can_trade=True):
            bot: "Bot"
            orders = [*orders, bot.give_order(closed_candle, current_candle)]
        return orders

    @staticmethod
    def get_asset_price(exchange_account, base: str, quote: str = "USDT") -> float:
        """Get the price of an asset."""
        if base in STABLECOINS[exchange_account.exchange.name]:
            return 1
        if base + quote not in EXCHANGE_PAIRS[exchange_account.exchange.name]:
            error_msg = f"Invalid pair: {base+quote} on {exchange_account.exchange.name}"
            raise ControllerError.InvalidPair(error_msg)
        controller = Controller.get(exchange_account=exchange_account, base=base, quote=quote, interval="1m")

        return float(controller.get_price())

    def get_price_always(self) -> float:
        """Get the price of the pair.

        Always calls the exchange API. (Can be costly)

        Returns:
        float: The price of the pair.
        """
        exchange_controller = self.exchange_controller
        if exchange_controller.__class__ == BinanceController:
            try:
                self.price = float(exchange_controller.client.get_ticker(symbol=self.pair)["lastPrice"])
                self.last_price_update = datetime.now(tz=timezone.utc)
            except ReadTimeout:
                print("ReadTimeout in get_price_always")
            except SSLError:
                print("SSLError in get_price_always")
            except ConnectionError:
                print("ConnectionError in get_price_always")
        else:
            error_msg = f"Exchange controller not supported: {exchange_controller.__class__}"
            raise NotImplementedError(error_msg)
        self.save()
        return self.price

    def get_price(self) -> float:
        """Retreive the price of the pair.

        Only updates the price if it is older than 1 minute.
        The ControllerUpdate task automaticaly refreshes price data.

        Returns:
            price: The price of the pair.
        """
        if self.last_price_update is None or self.last_price_update < datetime.now(tz=timezone.utc) - timedelta(minutes=1):
            self.get_price_always()
        return self.price

    def download(
        self,
        start_date: datetime,
        end_date: datetime,
        verbose: int = 0,
        *,
        squash: bool = False,
    ):
        return self.exchange_controller.download(
            controller=self,
            start_date=start_date,
            end_date=end_date,
            squash=squash,
            verbose=verbose,
        )



from django_napse.utils.constants import MODIFICATION_STATUS
from django_napse.utils.usefull_functions import process_value_from_type

from .modification import Modification


class StrategyModification(Modification):
    def apply(self):
        strategy = self.order.connection.bot.strategy.find()
        strategy, self = self.apply__no_db(strategy)
        strategy.save()
        self.save()

    def apply__no_db(self, **kwargs):
        strategy = kwargs.get("strategy")
        if not hasattr(strategy, f"variable_{self.key}"):
            error_msg: str = f"Strategy {strategy} must have attribute variable_{self.key}"
            raise ValueError(error_msg)
        setattr(strategy, f"variable_{self.key}", process_value_from_type(self.value, self.target_type))
        self.status = MODIFICATION_STATUS.APPLIED
        return strategy, self



from typing import TYPE_CHECKING

from django.apps import apps
from django.db import models

from django_napse.core.models.histories.wallet import WalletHistory
from django_napse.utils.errors import WalletError

if TYPE_CHECKING:
    from django_napse.core.models.wallets.wallet import Wallet


class WalletManager(models.Manager):
    """Manager for the Wallet model."""

    def create(self, title: str, owner: models.Model) -> "Wallet":
        """Create a Wallet object.

        Args:
            title (str): The title of the wallet.
            owner (models.Model): The owner of the wallet.

        Raises:
            WalletError.CreateError: If the model is a Wallet (you should use a subclass instead.)

        Returns:
            Wallet: The created wallet.
        """
        if self.model == apps.get_model("django_napse_core", "Wallet"):
            error_msg = "WalletManager cannot create a Wallet object. Use a subclass instead."
            raise WalletError.CreateError(error_msg)
        wallet = self.model(title=title, owner=owner)
        wallet.save()
        WalletHistory.get_or_create(owner=wallet)
        return wallet



from django.apps import apps
from django.db import models


class SpaceManager(models.Manager):
    def create(self, name: str, exchange_account, description: str = ""):
        """Create a Space instance."""
        SpaceWallet = apps.get_model("django_napse_core", "SpaceWallet")
        SpaceSimulationWallet = apps.get_model("django_napse_core", "SpaceSimulationWallet")

        space = self.model(
            name=name,
            exchange_account=exchange_account,
            description=description,
        )
        space.save()

        SpaceWallet.objects.create(owner=space, title=f"Wallet for space {space.name}")
        SpaceSimulationWallet.objects.create(owner=space, title=f"Simulation Wallet for space {space.name}")

        return space



from django.db import models

from django_napse.core.models.fleets.cluster import Cluster
from django_napse.utils.errors import FleetError


class FleetManager(models.Manager):
    def create(
        self,
        name,
        exchange_account,
        clusters=None,
    ):
        clusters = clusters or []

        if sum(cluster["share"] for cluster in clusters) != 1:
            error_message = "The sum of all shares must be 1."
            raise FleetError.InvalidShares(error_message)

        if len(clusters) == 0:
            error_msg = "A fleet must have at least one cluster."
            raise FleetError.InvalidClusters(error_msg)

        fleet = self.model(
            name=name,
            exchange_account=exchange_account,
        )

        fleet.save()

        for cluster in clusters:
            cluster["template_bot"] = cluster["template_bot"].copy()
            Cluster.objects.create(fleet=fleet, **cluster)

        fleet.setup_finished = True
        fleet.save()
        return fleet
